import cv2
import pandas as pd
import numpy as np
from pathlib import Path
from typing import Tuple, Union, List
from torch.utils.data import Dataset, DataLoader
import torch
from torch import nn
from concurrent.futures import ThreadPoolExecutor, as_completed
from collections import namedtuple
from transformers import OneFormerProcessor, OneFormerForUniversalSegmentation
from transformers import AutoImageProcessor, Mask2FormerForUniversalSegmentation
import os
from tqdm.auto import tqdm
import json
from collections import defaultdict
from tqdm.contrib.concurrent import thread_map
import glob
import shutil

# a label and all meta information
Label = namedtuple( 'Label' , [

            'name'        , # The identifier of this label, e.g. 'car', 'person', ... .
                            # We use them to uniquely name a class

            'id'          , # An integer ID that is associated with this label.
                            # The IDs are used to represent the label in ground truth images
                            # An ID of -1 means that this label does not have an ID and thus
                            # is ignored when creating ground truth images (e.g. license plate).
                            # Do not modify these IDs, since exactly these IDs are expected by the
                            # evaluation server.

            'trainId'     , # Feel free to modify these IDs as suitable for your method. Then create
                            # ground truth images with train IDs, using the tools provided in the
                            # 'preparation' folder. However, make sure to validate or submit results
                            # to our evaluation server using the regular IDs above!
                            # For trainIds, multiple labels might have the same ID. Then, these labels
                            # are mapped to the same class in the ground truth images. For the inverse
                            # mapping, we use the label that is defined first in the list below.
                            # For example, mapping all void-type classes to the same ID in training,
                            # might make sense for some approaches.
                            # Max value is 255!

            'category'    , # The name of the category that this label belongs to

            'categoryId'  , # The ID of this category. Used to create ground truth images
                            # on category level.

            'hasInstances', # Whether this label distinguishes between single instances or not

            'ignoreInEval', # Whether pixels having this class as ground truth label are ignored
                            # during evaluations or not

            'color'       , # The color of this label
            ] )
        
def create_cityscapes_label_colormap():
        """Creates a label colormap used in CITYSCAPES segmentation benchmark.
        Returns:
            A colormap for visualizing segmentation results.
        """
        
        labels = [
        #       name                     id    trainId   category            catId     hasInstances   ignoreInEval   color
        Label(  'unlabeled'            ,  0 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'ego vehicle'          ,  1 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'rectification border' ,  2 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'out of roi'           ,  3 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'static'               ,  4 ,      255 , 'void'            , 0       , False        , True         , (  0,  0,  0) ),
        Label(  'dynamic'              ,  5 ,      255 , 'void'            , 0       , False        , True         , (111, 74,  0) ),
        Label(  'ground'               ,  6 ,      255 , 'void'            , 0       , False        , True         , ( 81,  0, 81) ),
        Label(  'road'                 ,  7 ,        0 , 'flat'            , 1       , False        , False        , (128, 64,128) ),
        Label(  'sidewalk'             ,  8 ,        1 , 'flat'            , 1       , False        , False        , (244, 35,232) ),
        Label(  'parking'              ,  9 ,      255 , 'flat'            , 1       , False        , True         , (250,170,160) ),
        Label(  'rail track'           , 10 ,      255 , 'flat'            , 1       , False        , True         , (230,150,140) ),
        Label(  'building'             , 11 ,        2 , 'construction'    , 2       , False        , False        , ( 70, 70, 70) ),
        Label(  'wall'                 , 12 ,        3 , 'construction'    , 2       , False        , False        , (102,102,156) ),
        Label(  'fence'                , 13 ,        4 , 'construction'    , 2       , False        , False        , (190,153,153) ),
        Label(  'guard rail'           , 14 ,      255 , 'construction'    , 2       , False        , True         , (180,165,180) ),
        Label(  'bridge'               , 15 ,      255 , 'construction'    , 2       , False        , True         , (150,100,100) ),
        Label(  'tunnel'               , 16 ,      255 , 'construction'    , 2       , False        , True         , (150,120, 90) ),
        Label(  'pole'                 , 17 ,        5 , 'object'          , 3       , False        , False        , (153,153,153) ),
        Label(  'polegroup'            , 18 ,      255 , 'object'          , 3       , False        , True         , (153,153,153) ),
        Label(  'traffic light'        , 19 ,        6 , 'object'          , 3       , False        , False        , (250,170, 30) ),
        Label(  'traffic sign'         , 20 ,        7 , 'object'          , 3       , False        , False        , (220,220,  0) ),
        Label(  'vegetation'           , 21 ,        8 , 'nature'          , 4       , False        , False        , (107,142, 35) ),
        Label(  'terrain'              , 22 ,        9 , 'nature'          , 4       , False        , False        , (152,251,152) ),
        Label(  'sky'                  , 23 ,       10 , 'sky'             , 5       , False        , False        , ( 70,130,180) ),
        Label(  'person'               , 24 ,       11 , 'human'           , 6       , True         , False        , (220, 20, 60) ),
        Label(  'rider'                , 25 ,       12 , 'human'           , 6       , True         , False        , (255,  0,  0) ),
        Label(  'car'                  , 26 ,       13 , 'vehicle'         , 7       , True         , False        , (  0,  0,142) ),
        Label(  'truck'                , 27 ,       14 , 'vehicle'         , 7       , True         , False        , (  0,  0, 70) ),
        Label(  'bus'                  , 28 ,       15 , 'vehicle'         , 7       , True         , False        , (  0, 60,100) ),
        Label(  'caravan'              , 29 ,      255 , 'vehicle'         , 7       , True         , True         , (  0,  0, 90) ),
        Label(  'trailer'              , 30 ,      255 , 'vehicle'         , 7       , True         , True         , (  0,  0,110) ),
        Label(  'train'                , 31 ,       16 , 'vehicle'         , 7       , True         , False        , (  0, 80,100) ),
        Label(  'motorcycle'           , 32 ,       17 , 'vehicle'         , 7       , True         , False        , (  0,  0,230) ),
        Label(  'bicycle'              , 33 ,       18 , 'vehicle'         , 7       , True         , False        , (119, 11, 32) ),
        Label(  'license plate'        , -1 ,       -1 , 'vehicle'         , 7       , False        , True         , (  0,  0,142) ),
        ]
        return labels

def create_mapillary_vistas_label_colormap():
    """Creates a label colormap used in Mapillary Vistas segmentation benchmark.
    Returns:
        A list of labels for visualizing segmentation results.
    """

    labels = [
        Label('Bird',                   0,      0, 'animal',             0,      True,         False,        (165, 42, 42)),
        Label('Ground Animal',          1,      1, 'animal',             0,      True,         False,        (0, 192, 0)),
        Label('Curb',                   2,      2, 'construction',       1,      False,        False,        (196, 196, 196)),
        Label('Fence',                  3,      3, 'construction',       1,      False,        False,        (190, 153, 153)),
        Label('Guard Rail',             4,      4, 'construction',       1,      False,        False,        (180, 165, 180)),
        Label('Barrier',                5,      5, 'construction',       1,      False,        False,        (102, 102, 156)),
        Label('Wall',                   6,      6, 'construction',       1,      False,        False,        (102, 102, 156)),
        Label('Bike Lane',              7,      7, 'flat',               2,      False,        False,        (128, 64, 255)),
        Label('Crosswalk - Plain',      8,      8, 'flat',               2,      False,        False,        (140, 140, 200)),
        Label('Curb Cut',               9,      9, 'flat',               2,      False,        False,        (170, 170, 170)),
        Label('Parking',               10,     10, 'flat',               2,      False,        False,        (250, 170, 160)),
        Label('Pedestrian Area',       11,     11, 'flat',               2,      False,        False,        (96, 96, 96)),
        Label('Rail Track',            12,     12, 'flat',               2,      False,        False,        (230, 150, 140)),
        Label('Road',                  13,     13, 'flat',               2,      False,        False,        (128, 64, 128)),
        Label('Service Lane',          14,     14, 'flat',               2,      False,        False,        (110, 110, 110)),
        Label('Sidewalk',              15,     15, 'flat',               2,      False,        False,        (244, 35, 232)),
        Label('Bridge',                16,     16, 'construction',       1,      False,        False,        (150, 100, 100)),
        Label('Building',              17,     17, 'construction',       1,      False,        False,        (70, 70, 70)),
        Label('Tunnel',                18,     18, 'construction',       1,      False,        False,        (150, 120, 90)),
        Label('Person',                19,     19, 'human',              3,      True,         False,        (220, 20, 60)),
        Label('Bicyclist',             20,     20, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Motorcyclist',          21,     21, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Other Rider',           22,     22, 'human',              3,      True,         False,        (255, 0, 0)),
        Label('Lane Marking - Crosswalk',23,  23, 'marking',            4,      False,        True,         (200, 128, 128)),
        Label('Lane Marking - General',24,     24, 'marking',            4,      True,         False,        (255, 255, 255)),
        Label('Mountain',              25,     25, 'nature',             5,      False,        False,        (64, 170, 64)),
        Label('Sand',                  26,     26, 'nature',             5,      False,        False,        (230, 160, 50)),
        Label('Sky',                   27,     27, 'sky',                6,      False,        False,        (70, 130, 180)),
        Label('Snow',                  28,     28, 'nature',             5,      False,        False,        (190, 255, 255)),
        Label('Terrain',               29,     29, 'nature',             5,      False,        False,        (152, 251, 152)),
        Label('Vegetation',            30,     30, 'nature',             5,      False,        False,        (107, 142, 35)),
        Label('Water',                 31,     31, 'water',              7,      False,        False,        (0, 170, 30)),
        Label('Banner',                32,     32, 'object',             8,      False,        False,        (255, 220, 0)),
        Label('Bench',                 33,     33, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Bike Rack',             34,     34, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Billboard',             35,     35, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Catch Basin',           36,     36, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('CCTV Camera',           37,     37, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Fire Hydrant',          38,     38, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Junction Box',          39,     39, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Mailbox',               40,     40, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Manhole',               41,     41, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Phone Booth',           42,     42, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Pothole',               43,     43, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Street Light',          44,     44, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Pole',                  45,     45, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Traffic Sign Frame',    46,     46, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Utility Pole',          47,     47, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Traffic Light',         48,     48, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Traffic Sign (Back)',   49,     49, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Traffic Sign (Front)',  50,     50, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Trash Can',             51,     51, 'object',             8,      False,        False,        (255, 0, 0)),
        Label('Bicycle',               52,     52, 'vehicle',            9,      True,         False,        (119, 11, 32)),
        Label('Boat',                  53,     53, 'vehicle',            9,      False,        False,        (0, 0, 142)),
        Label('Bus',                   54,     54, 'vehicle',            9,      True,         False,        (0, 60, 100)),
        Label('Car',                   55,     55, 'vehicle',            9,      True,         False,        (0, 0, 142)),
        Label('Caravan',               56,     56, 'vehicle',            9,      True,         False,        (0, 0, 90)),
        Label('Motorcycle',            57,     57, 'vehicle',            9,      True,         False,        (0, 0, 230)),
        Label('On Rails',              58,     58, 'vehicle',            9,      False,        False,        (0, 80, 100)),
        Label('Other Vehicle',         59,     59, 'vehicle',            9,      True,         False,        (128, 64, 64)),
        Label('Trailer',               60,     60, 'vehicle',            9,      True,         False,        (0, 0, 110)),
        Label('Truck',                 61,     61, 'vehicle',            9,      True,         False,        (0, 0, 70)),
        Label('Wheeled Slow',          62,     62, 'vehicle',            9,      False,        False,        (0, 0, 192)),
        Label('Car Mount',             63,     63, 'vehicle',            9,      True,         False,        (32, 32, 32)),
        Label('Ego Vehicle',           64,     64, 'vehicle',            9,      True,         False,        (120, 10, 10))
    ]
    return labels

def get_resized_dimensions(image: cv2.Mat, max_size: int = 2048) -> Tuple[int, int]:
    """
    Get the resized dimensions of an image based on a maximum size.

    Args:
        image (cv2.Mat): Input image in OpenCV format.
        max_size (int, optional): Maximum size for the larger side of the image. Defaults to 2048.

    Returns:
        Tuple[int, int]: Tuple containing new height and width.
    """
    height, width = image.shape[:2]

    # Determine the larger side
    larger_side = max(height, width)

    # Check if the larger side exceeds the maximum size
    if larger_side > max_size:
        # Calculate the scaling factor
        scaling_factor = max_size / larger_side

        # Calculate the new dimensions
        new_width = int(width * scaling_factor)
        new_height = int(height * scaling_factor)

        return (new_height, new_width)

    return (height, width)

class ImageDataset(Dataset):
    """
    Custom Dataset class for images.
    """

    def __init__(self, image_files: List[Path], rgb: bool = True) -> None:
        """
        Initialize the ImageDataset.

        Args:
            image_files (List[Path]): List of image files.
            rgb (bool, optional): Flag to convert images to RGB. Defaults to True.
        """
        # remove any non-image files and file names starting with "."
        image_files = [image_file for image_file in image_files if image_file.suffix.lower() in [".jpg", ".jpeg", ".png"] and not image_file.name.startswith(".")]
        self.image_files = image_files
        self.rgb = rgb

    def __len__(self) -> int:
        """
        Get the length of the dataset.

        Returns:
            int: Number of image files in the dataset.
        """
        return len(self.image_files)

    def __getitem__(self, idx: int) -> Tuple[str, cv2.Mat, Tuple[int, int]]:
        """
        Get an item from the dataset.

        Args:
            idx (int): Index of the image.

        Returns:
            Tuple[str, cv2.Mat, Tuple[int, int]]: Tuple containing image file path, image data, and original image dimensions.
        """
        image_file = self.image_files[idx]
        img = cv2.imread(str(image_file))
        
        if img is None:
            raise ValueError(f"Unable to read image at {image_file}")
        
        original_img_shape = get_resized_dimensions(img)
        
        if self.rgb:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            
        return str(image_file), img, original_img_shape

    def collate_fn(self, data: List[Tuple[str, cv2.Mat, Tuple[int, int]]]) -> Tuple[List[str], List[cv2.Mat], List[Tuple[int, int]]]:
        """
        Custom collate function for the dataset.

        Args:
            data (List[Tuple[str, cv2.Mat, Tuple[int, int]]]): List of tuples containing image file path, image data, and original image dimensions.

        Returns:
            Tuple[List[str], List[cv2.Mat], List[Tuple[int, int]]]: Tuple containing lists of image file paths, image data, and original image dimensions.
        """
        image_files, images, original_img_shape = zip(*data)
        return list(image_files), list(images), list(original_img_shape)

class Segmenter:
    def __init__(self, dataset="cityscapes", task = "semantic", device = None):
        """
        Initialize the Segmenter with a model and dataset.

        Args:
            model_name (str): The name of the pre-trained model.
            dataset (str): The name of the dataset.
        """
        self.device = self._get_device(device)
        self.dataset = dataset
        self.task = task
        self.model_name = self._get_model_name(self.dataset, self.task)
        self.model, self.processor = self._get_model_processor(self.model_name)
        self.color_map = self._create_color_map(dataset)
        self.label_map = self._create_label_map(dataset)
        self.id_to_name_map = self._create_id_to_name_map(dataset)
    
    def _get_model_name(self, dataset, task):
        if dataset == "cityscapes":
            if task == "semantic":
                return "facebook/mask2former-swin-tiny-cityscapes-semantic"
            elif task == "panoptic":
                return "facebook/mask2former-swin-tiny-cityscapes-panoptic"
        elif dataset == "mapillary":
            if task == "semantic":
                return "facebook/mask2former-swin-large-mapillary-vistas-semantic"
            elif task == "panoptic":
                return "facebook/mask2former-swin-large-mapillary-vistas-panoptic"
        else:
            raise ValueError(f"Unknown dataset: {dataset}")

    def _get_model_processor(self, model_name):
        """
        Get the model and processor for the given model name.

        Args:
            model_name (str): The name of the pre-trained model.

        Returns:
            Tuple: The model and processor.
        """
        # Add other models in the future
        if "mask2former" in model_name:
            processor = AutoImageProcessor.from_pretrained(model_name)
            model = Mask2FormerForUniversalSegmentation.from_pretrained(model_name).to(self.device)
        return model, processor

    def _create_color_map(self, dataset):
        """
        Create a color map based on the given dataset.

        Args:
            dataset (str): The name of the dataset.

        Returns:
            np.ndarray: A color map.
        """
        if dataset == "cityscapes":
            labels = create_cityscapes_label_colormap()
        elif dataset == "mapillary":
            labels = create_mapillary_vistas_label_colormap()
        else:
            raise ValueError(f"Unknown dataset: {dataset}")

        labels = [label for label in labels if label.trainId != -1]
        train_ids = np.array([label.trainId for label in labels], dtype=np.uint8)
        colors = np.array([label.color for label in labels], dtype=np.uint8)
        max_train_id = np.max(train_ids) + 1
        color_map = np.zeros((max_train_id, 3), dtype=np.uint8)
        color_map[train_ids] = colors

        # Add a train_id_to_name dictionary as an attribute
        self.train_id_to_name = {label.trainId: label.name for label in labels}

        return color_map

    def _create_label_map(self, dataset):
        """
        Create a label map based on the given dataset.

        Args:
            dataset (str): The name of the dataset.

        Returns:
            Dict[Tuple, Label]: A dictionary mapping colors to labels.
        """
        if dataset == "cityscapes":
            labels = create_cityscapes_label_colormap()
        elif dataset == "mapillary":
            labels = create_mapillary_vistas_label_colormap()
        else:
            raise ValueError(f"Unknown dataset: {dataset}")

        color_to_label = {}
        for label in labels:
            color = label.color
            color_to_label[color] = label

        return color_to_label

    def _create_id_to_name_map(self, dataset):
        if dataset == "cityscapes":
            labels = create_cityscapes_label_colormap()
        elif dataset == "mapillary":
            labels = create_mapillary_vistas_label_colormap()
        return {label.trainId: label.name for label in labels}
    
    def _get_device(self, device) -> torch.device:
        """
        Get the appropriate device for running the model.

        Returns:
            torch.device: The device to use for running the model.
        """
        if device is not None:
            if device not in ["cpu", "cuda", "mps"]:
                raise ValueError(f"Unknown device: {device}")
            else:
                print(f"Using {device.upper()}")
                return torch.device(device)
        if torch.cuda.is_available():
            print("Using GPU")
            return torch.device("cuda")
        else:
            print("Using CPU")
            return torch.device("cpu")

    def _calculate_pixel_ratios(self, segmented_img):
        """
        Calculate pixel ratios for each class in the segmented image.

        Args:
            segmented_img (numpy.ndarray): Segmented image.

        Returns:
            dict: A dictionary with class names as keys and pixel ratios as values.
        """
        unique, counts = np.unique(segmented_img, return_counts=True)
        total_pixels = np.sum(counts)
        pixel_ratios = {self.train_id_to_name[train_id]: count / total_pixels for train_id, count in zip(unique, counts)}

        return pixel_ratios

    def _save_as_csv(self, input_dict, dir_output, value_name, csv_format):
        if csv_format == "long":
            df_list = [pd.DataFrame({'filename_key': [filename_key], 'label_name': [key], value_name: [value] if value is not None else [0]}) 
                    for filename_key, inner_dict in input_dict.items() 
                    for key, value in inner_dict.items()]

            pixel_ratios_df = pd.concat(df_list, ignore_index=True)

        elif csv_format == "wide":
            pixel_ratios_df = pd.DataFrame(input_dict).transpose().fillna(0)
            pixel_ratios_df.index.names = ["filename_key"]

        pixel_ratios_df.to_csv(dir_output / Path(value_name + ".csv"))


    def _panoptic_segmentation(self, images, original_img_shape):
        """
        Perform panoptic segmentation on the given images.

        Args:
            images (list): List of input images.
            original_img_shape (tuple): Original image shape.

        Returns:
            list: List of panoptic segmentation outputs.
        """
        inputs = self.processor(images=images, task_inputs=["panoptic"], return_tensors="pt").to(self.model.device)
        outputs = self.model(**inputs)
        return self.processor.post_process_panoptic_segmentation(outputs, target_sizes=original_img_shape, label_ids_to_fuse = set([]))

    def _semantic_segmentation(self, images, original_img_shape):
        """
        Perform semantic segmentation on the given images.

        Args:
            images (list): List of input images.
            original_img_shape (tuple): Original image shape.

        Returns:
            tuple: Tuple containing list of semantic segmentation outputs and list of pixel ratios.
        """
        inputs = self.processor(images=images, return_tensors="pt").to(self.model.device)
        with torch.no_grad():
            outputs = self.model(**inputs)
        segmentations = self.processor.post_process_semantic_segmentation(outputs, target_sizes=original_img_shape)
        return segmentations

    def _trainid_to_color(self, segmented_img):
        """
        Convert segmented image with train IDs to a colored image.

        Args:
            segmented_img (numpy.ndarray): Segmented image with train IDs.

        Returns:
            numpy.ndarray: Colored segmented image.
        """
        colored_img = self.color_map[segmented_img]
        return colored_img

    def _save_panoptic_segmentation_image(self, image_file, img, dir_output, output):
        """
        Save the panoptic segmentation image as a blended image with the original input image.

        Args:
            image_file (str): The input image file path.
            img (np.array): The input image in the format of a NumPy array.
            dir_output (Path): The output directory path to save the blended image.
            output (dict): The output dictionary containing the segmentation data.
        """
        colored_segmented_img = self._trainid_to_color(output['label_segmentation'].cpu().numpy())
        alpha = 0.5
        blend_img = cv2.addWeighted(img, alpha, colored_segmented_img, 1 - alpha, 0)

        # Add annotations for each segment
        for segment_info in output['segments_info']:
            segment_id = segment_info['id']
            label_id = segment_info['label_id']
            score = segment_info['score']
            
            # Use the label name instead of the label_id
            label_name = self.id_to_name_map.get(label_id)
            
            # Find the center of the segment for the label placement
            y, x = np.where(output['segmentation'].cpu().numpy() == segment_id)
            center_x, center_y = np.mean(x), np.mean(y)
            
            # Add the annotation
            text = f"{label_name}-{score:.2f}"
            cv2.putText(blend_img, text, (int(center_x), int(center_y)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1, cv2.LINE_AA)

        output_file = dir_output / Path(image_file).name
        
        # Save images based on specified options
        if "segmented_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_colored_segmented.png")), cv2.cvtColor(colored_segmented_img, cv2.COLOR_RGB2BGR))
        if "blend_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_blend.png")), cv2.cvtColor(blend_img, cv2.COLOR_RGB2BGR))
        
    def _save_semantic_segmentation_image(self, image_file, img, dir_output, output):
        """
        Save the semantic segmentation image as a colored segmented image and/or a blended image with the original input image.

        Args:
            image_file (str): The input image file path.
            img (np.array): The input image in the format of a NumPy array.
            dir_output (Path): The output directory path to save the colored segmented and/or blended image.
            output (Tensor): The output tensor containing the semantic segmentation data.
        """
        colored_segmented_img = self._trainid_to_color(output.cpu().numpy())
        alpha = 0.5
        blend_img = cv2.addWeighted(img, alpha, colored_segmented_img, 1 - alpha, 0)

        output_file = dir_output / Path(image_file).name
        
        # Save images based on specified options
        if "segmented_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_colored_segmented.png")), cv2.cvtColor(colored_segmented_img, cv2.COLOR_RGB2BGR))
        if "blend_image" in self.save_image_options:
            cv2.imwrite(str(output_file.with_name(output_file.stem + "_blend.png")), cv2.cvtColor(blend_img, cv2.COLOR_RGB2BGR))

    def _panoptic_count_labels(self, output):
        label_counts = {}

        # Loop through each segment in the image
        segments_info_list = output['segments_info']
        for segments_info in segments_info_list:
            # Convert label_id to label_name
            label_name = self.id_to_name_map.get(segments_info["label_id"])
            # Increment the count for the name in label_counts
            if label_name in label_counts:
                label_counts[label_name] += 1
            else:
                label_counts[label_name] = 1
        return label_counts
    
    def _panoptic_segment_to_label(self, output):
        """
        This function converts the output of post_process_panoptic_segmentation function
        from segment_id to label_id.
        :param output: The output dictionary from the post_process_panoptic_segmentation function
        :return: segmentation with label_ids instead of segment_ids
        """
        # Extract the segmentation and segments_info from the output
        segmentation = output['segmentation']
        segments_info = output['segments_info']

        # Create a mapping from segment_id to label_id
        id_map = {segment['id']: segment['label_id'] for segment in segments_info}

        # Use the map to convert the segmentation tensor from segment_ids to label_ids
        new_segmentation = segmentation.clone()

        for seg_id, label_id in id_map.items():
            new_segmentation[segmentation == seg_id] = label_id

        return new_segmentation

    def _process_images(self, task, image_files, images, dir_output, pixel_ratio_dict, original_img_shape, panoptic_dict=None):
        """
        Process the input images for segmentation and save the output images.

        Args:
            task (str): The segmentation task to perform, either "panoptic" or "semantic".
            image_files (List[str]): The list of file paths of the input images.
            images (List[ndarray]): The list of input images in the form of numpy arrays.
            dir_output (Path): The output directory where the segmented images will be saved.
            pixel_ratio_dict (defaultdict): A dictionary to store the pixel ratios of the segmented images.
            original_img_shape (List[Tuple[int, int]]): The original shapes of the input images.

        Returns:
            None
        """
        outputs = None
        if task == "panoptic":
            outputs = self._panoptic_segmentation(images, original_img_shape)
            if outputs is not None:
                for image_file, img, output in zip(image_files, images, outputs):
                    # create a new segmentation with label_ids instead of segment_ids
                    output["label_segmentation"] = self._panoptic_segment_to_label(output)
                    if (len(self.save_image_options) > 0) & (dir_output is not None):
                        self._save_panoptic_segmentation_image(image_file, img, dir_output, output)
                    pixel_ratio = self._calculate_pixel_ratios(output["label_segmentation"].cpu().numpy()) 
                    label_counts = self._panoptic_count_labels(output)
                    image_file_key = Path(image_file).stem
                    pixel_ratio_dict[image_file_key] = pixel_ratio
                    panoptic_dict[image_file_key] = label_counts
                    
        elif task == "semantic":
            segmentations = self._semantic_segmentation(images, original_img_shape)
            if segmentations is not None:
                for image_file, img, segmentation in zip(image_files, images, segmentations):
                    if (len(self.save_image_options) > 0) & (dir_output is not None):
                        self._save_semantic_segmentation_image(image_file, img, dir_output, segmentation)
                    pixel_ratio = self._calculate_pixel_ratios(segmentation.cpu().numpy())
                    image_file_key = Path(image_file).stem
                    pixel_ratio_dict[image_file_key] = pixel_ratio

    # Modify the segment method inside the Segmenter class
    def segment(self, dir_input: Union[str, Path], 
                dir_image_output: Union[str, Path, None] = None, 
                dir_segmentation_summary_output: Union[str, Path, None] = None, 
                batch_size=1, 
                save_image_options = ["segmented_image", "blend_image"], 
                pixel_ratio_save_format = ["json", "csv"],
                csv_format = "long", # "long" or "wide"
                max_workers: Union[int, None] = None):
        # make sure that at least one of dir_image_output and dir_segmentation_summary_output is not None
        if (dir_image_output is None) & (dir_segmentation_summary_output is None):
            raise ValueError("At least one of dir_image_output and dir_segmentation_summary_output must not be None.")
        
        # save_image_options as a property of the class
        self.save_image_options = save_image_options
        
        # make directory
        dir_input = Path(dir_input)

        # initialize completed_image_files
        completed_image_files = set()
        if dir_image_output is not None:
            dir_image_output = Path(dir_image_output)
            dir_image_output.mkdir(parents=True, exist_ok=True)
            # get a list of .png files and _blend.png files in the output directory and get the file names as a set
            completed_image_files.update([str(Path(f).stem).replace("_blend", "") for f in dir_image_output.glob('*.png')])
                        
        if dir_segmentation_summary_output is not None:
            dir_segmentation_summary_output = Path(dir_segmentation_summary_output)
            dir_segmentation_summary_output.mkdir(parents=True, exist_ok=True)
            # Create a new directory called "pixel_ratio_checkpoints"
            dir_cache_segmentation_summary = dir_segmentation_summary_output / 'pixel_ratio_checkpoints'
            dir_cache_segmentation_summary.mkdir(parents=True, exist_ok=True)

            # Load all the checkpoint json files
            checkpoints = glob.glob(str(dir_cache_segmentation_summary / '*.json'))
            checkpoint_start_index = len(checkpoints)

            if checkpoint_start_index > 0:
                for checkpoint in checkpoints:
                    with open(checkpoint, 'r') as f:
                        checkpoint_dict = json.load(f)
                        completed_image_files.update(checkpoint_dict.keys())
            
            # also check pixel_ratios.json in dir_cache_segmentation_summary
            if (dir_cache_segmentation_summary / 'pixel_ratios.json').exists():
                with open(dir_cache_segmentation_summary / 'pixel_ratios.json', 'r') as f:
                    pixel_ratio_dict = json.load(f)
                    completed_image_files.update(pixel_ratio_dict.keys())

        # Get the list of all image files and filter the ones that are not completed yet
        # List of possible image file extensions
        image_extensions = [".jpg", ".jpeg", ".png", ".tif", ".tiff", ".bmp", ".dib", ".pbm", ".pgm", ".ppm", ".sr", ".ras", ".exr", ".jp2"]

        # Get the list of all image files in the directory that are not completed yet
        image_file_list = [f for f in Path(dir_input).iterdir() if f.suffix in image_extensions and f.stem not in completed_image_files]

        outer_batch_size = 1000  # Number of inner batches in one outer batch
        num_outer_batches = (len(image_file_list) + outer_batch_size * batch_size - 1) // (outer_batch_size * batch_size)

        for i in tqdm(range(num_outer_batches), desc=f"Processing outer batches of size {min(outer_batch_size * batch_size, len(image_file_list))}"):
            # Get the image files for the current outer batch
            outer_batch_image_file_list = image_file_list[i * outer_batch_size * batch_size : (i+1) * outer_batch_size * batch_size]

            dataset = ImageDataset(outer_batch_image_file_list)
            dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=dataset.collate_fn)

            # set up pixel_ratio_dict for the current outer batch
            pixel_ratio_dict = defaultdict(dict)  # reset pixel_ratio_dict for each outer batch
            panoptic_dict = defaultdict(dict)  # reset panoptic_dict for each outer batch
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = []

                for batch in dataloader:
                    image_files, images, original_img_shape = batch
                    if self.task == "panoptic":
                        future = executor.submit(self._process_images, self.task, image_files, images, dir_image_output, pixel_ratio_dict, original_img_shape, panoptic_dict)
                    elif self.task == "semantic":
                        future = executor.submit(self._process_images, self.task, image_files, images, dir_image_output, pixel_ratio_dict, original_img_shape)
                    futures.append(future)

                for completed_future in tqdm(as_completed(futures), total=len(futures), desc=f"Processing outer batch #{i+1}"):
                    completed_future.result()

                if dir_segmentation_summary_output is not None:
                    # Save checkpoint for each outer batch
                    with open(f'{dir_cache_segmentation_summary}/checkpoint_batch_{checkpoint_start_index+i+1}_pixel_ratio.json', 'w') as f:
                        json.dump(pixel_ratio_dict, f)
                        
                    if self.task == "panoptic":
                        with open(f'{dir_cache_segmentation_summary}/checkpoint_batch_{checkpoint_start_index+i+1}_panoptic.json', 'w') as f:
                            json.dump(panoptic_dict, f)
        if dir_segmentation_summary_output is not None:
            # Merge all checkpoints into a single pixel_ratio_dict
            pixel_ratio_dict = defaultdict(dict)
            for checkpoint in glob.glob(str(dir_cache_segmentation_summary / '*_pixel_ratio.json')):
                with open(checkpoint, 'r') as f:
                    checkpoint_dict = json.load(f)
                    for key, value in checkpoint_dict.items():
                        pixel_ratio_dict[key] = value
            
            # Merge all checkpoints into a single panoptic_dict
            if self.task == "panoptic":
                panoptic_dict = defaultdict(dict)
                for checkpoint in glob.glob(str(dir_cache_segmentation_summary / '*_panoptic.json')):
                    with open(checkpoint, 'r') as f:
                        checkpoint_dict = json.load(f)
                        for key, value in checkpoint_dict.items():
                            panoptic_dict[key] = value
                            
            # Merge existing pixel_ratios.json with the new pixel_ratio_dict
            if (dir_segmentation_summary_output / 'pixel_ratios.json').exists():
                with open(dir_segmentation_summary_output / 'pixel_ratios.json', 'r') as f:
                    existing_pixel_ratio_dict = json.load(f)
                    for key, value in existing_pixel_ratio_dict.items():
                        pixel_ratio_dict[key] = value
                        
            # Merge existing label_counts.json with the new panoptic_dict
            if self.task == "panoptic":
                if (dir_segmentation_summary_output / 'label_counts.json').exists():
                    with open(dir_segmentation_summary_output / 'label_counts.json', 'r') as f:
                        existing_panoptic_dict = json.load(f)
                        for key, value in existing_panoptic_dict.items():
                            panoptic_dict[key] = value

            # Save pixel_ratio_dict as a JSON or CSV file
            if "json" in pixel_ratio_save_format:
                with open(dir_segmentation_summary_output / "pixel_ratios.json", "w") as f:
                    json.dump(pixel_ratio_dict, f)
                if self.task == "panoptic":
                    with open(dir_segmentation_summary_output / "label_counts.json", "w") as f:
                        json.dump(panoptic_dict, f)
            if "csv" in pixel_ratio_save_format:
                self._save_as_csv(pixel_ratio_dict, dir_segmentation_summary_output, "pixel_ratios", csv_format)
                if self.task == "panoptic":
                    self._save_as_csv(panoptic_dict, dir_segmentation_summary_output, "label_counts", csv_format)
                    
            # Delete the "pixel_ratio_checkpoints" directory
            shutil.rmtree(dir_cache_segmentation_summary, ignore_errors=True)

            
    def calculate_pixel_ratio_post_process(self, dir_input, dir_output, pixel_ratio_save_format = ["json", "csv"]):
        """
        Calculates the pixel ratio of different classes present in the segmented images and saves the results in either JSON or CSV format.

        Args:
            dir_input: A string or Path object representing the input directory containing the segmented images.
            dir_output: A string or Path object representing the output directory where the pixel ratio results will be saved.
            pixel_ratio_save_format: A list containing the file formats in which the results will be saved. The allowed file formats are "json" and "csv". The default value is ["json", "csv"].

        Returns:
            None
        """
        def calculate_label_ratios(image, label_map):
            """
            Calculates the pixel ratio of different classes present in a single image.

            Args:
                image: A numpy array representing an image.
                label_map: A dictionary containing the label names and their respective RGB colors.

            Returns:
            A dictionary containing the pixel ratio of different classes in the given image.
            """
            label_ratios = {}
            total_pixels = image.shape[0] * image.shape[1]

            for color, label in label_map.items():
                color_pixels = np.count_nonzero(np.all(image == color, axis=-1))
                label_ratios[label.name] = color_pixels / total_pixels

            return label_ratios

        def process_image_file(image_file, label_map):
            """
            Calculates the pixel ratio of different classes in a single segmented image file.

            Args:
                image_file: A Path object representing the segmented image file.
                label_map: A dictionary containing the label names and their respective RGB colors.

            Returns:
                A tuple containing the image file key and the pixel ratio of different classes in the given image.
            """
            image_file_key = str(Path(image_file).stem).replace("_colored_segmented", "")
            image = cv2.imread(str(image_file))
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            label_ratios = calculate_label_ratios(image, label_map)
            return image_file_key, label_ratios

        def results_to_dataframe(results):
            """
            Converts the results obtained from processing each image file into a Pandas DataFrame.

            Args:
                results: A list of tuples, where each tuple contains the image file key and the pixel ratio of different classes in the corresponding image.

            Returns:
                A Pandas DataFrame containing the pixel ratios of different classes in each image file.
            """
            pixel_ratio_dict = {}

            for image_file_key, label_ratios in results:
                pixel_ratio_dict[str(image_file_key)] = label_ratios

            pixel_ratios_df = pd.DataFrame(pixel_ratio_dict).transpose()
            pixel_ratios_df.fillna(0, inplace=True)
            pixel_ratios_df.index.names = ["filename_key"]

            return pixel_ratios_df
        
        def results_to_nested_dict(results):
            """
            Converts the results obtained from processing each image file into a nested dictionary.

            Args:
                results: A list of tuples, where each tuple contains the image file key and the pixel ratio of different classes in the corresponding image.

            Returns:
                A nested dictionary containing the pixel ratios of different classes in each image file.
            """
            data = {}

            for image_file_key, label_ratios in results:
                image_file_key = str(image_file_key)
                data[image_file_key] = label_ratios

            return data

        # get files
        if isinstance(dir_input, str):
            dir_input = Path(dir_input)

        # Set image file extensions
        image_extensions = ['.jpg', '.png']

        image_files = [file for file in dir_input.rglob('*') if file.suffix.lower() in image_extensions and '_colored_segmented' in file.stem]

        results = thread_map(process_image_file, image_files, [self.label_map] * len(image_files))

        if "json" in pixel_ratio_save_format:
            json_output_file = Path(dir_output) / 'pixel_ratio.json'
            nested_dict = results_to_nested_dict(results)
            with open(json_output_file, 'w') as f:
                json.dump(nested_dict, f, indent=2) 
                
        if "csv" in pixel_ratio_save_format:
            csv_output_file = Path(dir_output) / 'pixel_ratio.csv'
            df = results_to_dataframe(results)
            df.to_csv(csv_output_file)


# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~

This package contains the class representations of logic within the Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import client  # noqa: F401
from . import exceptions  # noqa: F401
from . import geojson  # noqa: F401
from . import api  # noqa: F401
from . import logger  # noqa: F401
from . import config # noqa: F401



from abc import ABC, abstractmethod
import pkg_resources
import csv
import os
import pandas as pd

class BaseDownloader(ABC):
    @abstractmethod
    def __init__(self, log_path = None, distance = 1, grid = False, grid_size = 1):
        self._log_path = log_path
        self._distance = distance
        self._grid = grid
        self._grid_size = grid_size
        self._user_agents = self._get_ua()
        self._proxies = self._get_proxies()

    @property
    def log_path(self):
        return self._log_path    
    @log_path.setter
    def log_path(self,log_path):
        self._log_path = log_path
    
    @property
    def distance(self):
        return self._distance    
    @distance.setter
    def distance(self,distance):
        self._distance = distance
    
    @property
    def grid(self):
        return self._grid    
    @grid.setter
    def grid(self,grid):
        self._grid = grid
        
    @property
    def grid_size(self):
        return self._grid_size    
    @grid_size.setter
    def grid_size(self,grid_size):
        self._grid_size = grid_size

    @property
    def proxies(self):
        return self._proxies
    
    def _get_proxies(self):
        proxies_file = pkg_resources.resource_filename('zensvi.download.utils', 'proxies.csv')
        proxies = []
        # open with "utf-8" encoding to avoid UnicodeDecodeError
        with open(proxies_file, 'r', encoding = "utf-8") as f:
            reader = csv.DictReader(f)
            for row in reader:
                ip = row['ip']
                port = row['port']
                protocols = row['protocols']
                proxy_dict = {protocols: f"{ip}:{port}"}
                proxies.append(proxy_dict)
        return proxies
    
    @property
    def user_agents(self):
        return self._user_agents
    
    def _get_ua(self):
        user_agent_file = pkg_resources.resource_filename('zensvi.download.utils', 'UserAgent.csv')
        UA = []
        with open(user_agent_file, 'r') as f:
            for line in f:
                ua = {"user_agent": line.strip()}
                UA.append(ua)
        return UA
    
    def _log_write(self, pids):
        if self.log_path == None:
            return
        with open(self.log_path, 'a+') as fw:
            for pid in pids:
                fw.write(pid+'\n')
                
    def _check_already(self, all_panoids):
        # Get the set of already downloaded images
        name_r = set()
        for dirpath, dirnames, filenames in os.walk(self.panorama_output):
            for name in filenames:
                name_r.add(name.split(".")[0])

        # Filter the list of all panoids to only include those not already downloaded
        all_panoids = list(set(all_panoids) - name_r)
        return all_panoids
    
    def _read_pids(self, path_pid, start_date, end_date):
        pid_df = pd.read_csv(path_pid)
        # filter pids by date
        pid_df = self._filter_pids_date(pid_df, start_date, end_date)
        # get unique pids as a list
        pids = pid_df.iloc[:,0].unique().tolist()
        return pids

    @abstractmethod
    def _filter_pids_date(self, pid_df, start_date, end_date):
        pass

    @abstractmethod
    def download_svi(self, 
                    dir_output, 
                    lat=None, 
                    lon=None, 
                    input_csv_file="", 
                    input_shp_file = "", 
                    input_place_name = "", 
                    id_columns=None, 
                    buffer = 0,
                    update_pids = False,
                    start_date = None,
                    end_date = None,
                    metadata_only = False):
        pass



from shapely.geometry import shape, mapping, MultiPolygon, MultiLineString, MultiPoint

def convert_multi_feature_collection(feature_collection):
    """
    Convert all MultiPolygon, MultiLineString, or MultiPoint features in a GeoJSON FeatureCollection 
    to individual Polygon, LineString, or Point features, respectively.
    
    :param feature_collection: A GeoJSON FeatureCollection.
    :return: A new GeoJSON FeatureCollection with converted features.
    """
    converted_features = []

    for feature in feature_collection['features']:
        geometry_type = feature['geometry']['type']
        shapely_geometry = shape(feature['geometry'])

        if isinstance(shapely_geometry, (MultiPolygon, MultiLineString, MultiPoint)):
            for geom in shapely_geometry.geoms:
                singular_feature = {
                    'type': 'Feature',
                    'properties': feature['properties'],  # Copy the properties
                    'geometry': mapping(geom)
                }
                converted_features.append(singular_feature)
        else:
            converted_features.append(feature)  # If not a Multi type, return as is

    return {'type': 'FeatureCollection', 'features': converted_features}


def extract_coordinates_from_polygons(feature_collection):
    """
    Check if all features in the GeoJSON are Polygons or MultiPolygons.
    If so, extract their coordinates as a list of lists (each list contains tuples of longitude, latitude).
    Raise an error if any feature is not a Polygon or MultiPolygon.

    :param feature_collection: A GeoJSON FeatureCollection.
    :return: A list of lists with coordinates.
    """
    polygons_coordinates = []

    for feature in feature_collection['features']:
        geometry_type = feature['geometry']['type']
        shapely_geometry = shape(feature['geometry'])

        if geometry_type not in ['Polygon', 'MultiPolygon']:
            raise ValueError("All features must be Polygons or MultiPolygons. Found: " + geometry_type)

        if isinstance(shapely_geometry, MultiPolygon):
            for polygon in shapely_geometry.geoms:
                # Append each Polygon's exterior coordinates as a new list
                polygons_coordinates.append(list(polygon.exterior.coords))
        elif geometry_type == 'Polygon':  # It's a Polygon
            # Append the Polygon's exterior coordinates as a new list
            polygons_coordinates.append(list(shapely_geometry.exterior.coords))

    return polygons_coordinates

# Example usage:
# coordinates = extract_coordinates_from_polygons(your_feature_collection)




import random
import datetime
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed 
import warnings
import requests
from pathlib import Path
import geopandas as gpd
from tqdm import tqdm
from shapely.geometry import Point
import warnings
from shapely.errors import ShapelyDeprecationWarning
import math
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)

import glob
import shutil
import numpy as np
import osmnx as ox
from typing import List, Union

from zensvi.download.base import BaseDownloader
import zensvi.download.mapillary.interface as mly
from zensvi.download.utils.imtool import ImageTool
from zensvi.download.utils.get_pids import panoids
from zensvi.download.utils.geoprocess import GeoProcessor
from zensvi.download.utils.helpers import standardize_column_names, create_buffer_gdf

class GSVDownloader(BaseDownloader):
    def __init__(self, gsv_api_key: str = None, log_path: str = None, distance: int = 1, grid: bool = False, grid_size: int = 1):
        """
        Google Street View Downloader class.

        Args:
            gsv_api_key (str, optional): Google Street View API key. Defaults to None.
            log_path (str, optional): Path to the log file. Defaults to None.
            distance (int, optional): Distance parameter for the GeoProcessor. Defaults to 1.
            grid (bool, optional): Grid parameter for the GeoProcessor. Defaults to False.
            grid_size (int, optional): Grid size parameter for the GeoProcessor. Defaults to 1.

        Raises:
            Warning: If gsv_api_key is not provided.
        """
        super().__init__(log_path, distance, grid, grid_size)
        if gsv_api_key == None:
            warnings.warn("Please provide your Google Street View API key to augment metadata.")
        self._gsv_api_key = gsv_api_key

    @property
    def gsv_api_key(self) -> str:
        """
        Property to get the Google Street View API key.

        Returns:
            str: Google Street View API key.
        """
        return self._gsv_api_key    
    
    @gsv_api_key.setter
    def gsv_api_key(self, gsv_api_key: str):
        """
        Setter to set the Google Street View API key.

        Args:
            gsv_api_key (str): Google Street View API key.
        """
        self._gsv_api_key = gsv_api_key

    def _augment_metadata(self, df):
        if self.cache_pids_augmented.exists():
            df = pd.read_csv(self.cache_pids_augmented)
            print("The augmented panorama IDs have been read from the cache")
            return df
        
        # Create a new directory called "augmented_metadata_checkpoints"
        dir_cache_augmented_metadata = self.dir_cache / 'augmented_pids'
        dir_cache_augmented_metadata.mkdir(parents=True, exist_ok=True)

        # Load all the checkpoint csv files
        checkpoints = glob.glob(str(dir_cache_augmented_metadata / '*.csv'))
        checkpoint_start_index = len(checkpoints)

        if checkpoint_start_index > 0:
            completed_rows = pd.concat([pd.read_csv(checkpoint) for checkpoint in checkpoints], ignore_index=True)
            completed_indices = completed_rows.index.unique()

            # Filter df to get remaining indices to augment metadata for
            df = df.loc[~df.index.isin(completed_indices)]

        def get_year_month(pid, proxies):
            url = "https://maps.googleapis.com/maps/api/streetview/metadata?pano={}&key={}".format(pid, self.gsv_api_key)
            while True:
                proxy = random.choice(proxies)
                try:
                    response = requests.get(url, proxies=proxy, timeout=5)
                    break
                except Exception as e:
                    print(f"Proxy {proxy} is not working. Exception: {e}")
                    continue

            response = response.json()
            if response['status'] == 'OK':
                # get year and month from date
                try:
                    date = response['date']
                    year = date.split("-")[0]
                    month = date.split("-")[1]
                except Exception:
                    year = None
                    month = None
                return {"year": year, "month": month}
            return {"year": None, "month": None}    

        def worker(row, proxies):
            panoid = row.panoid
            year_month = get_year_month(panoid, proxies)
            return row.Index, year_month

        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size

        for i in tqdm(range(num_batches), desc=f"Augmenting metadata by batch size {min(batch_size, len(df))}"):
            batch_df = df.iloc[i*batch_size : (i+1)*batch_size].copy()  # Copy the batch data to a new dataframe
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row, self.proxies): row.Index for row in batch_df.itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Augmenting metadata for batch #{i+1}"):
                    row_index, year_month = future.result()
                    batch_df.at[row_index, 'year'] = year_month['year']
                    batch_df.at[row_index, 'month'] = year_month['month']

            # Save checkpoint for each batch
            batch_df.to_csv(f'{dir_cache_augmented_metadata}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
        
        # Merge all checkpoints into a single dataframe
        df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_augmented_metadata / '*.csv'))], ignore_index=True)

        # save the augmented metadata
        df.to_csv(self.cache_pids_augmented, index=False)
        # delete cache_lat_lon
        if self.cache_lat_lon.exists():
            self.cache_lat_lon.unlink() 
        # delete cache_pids_raw
        if self.cache_pids_raw.exists():
            self.cache_pids_raw.unlink()
        # delete the cache directory
        if dir_cache_augmented_metadata.exists():
            shutil.rmtree(dir_cache_augmented_metadata)
        return df

    def _get_pids_from_df(self, df, id_columns=None):
        # 1. Create a new directory called "pids" to store each batch pids
        dir_cache_pids = self.dir_cache / 'raw_pids'
        dir_cache_pids.mkdir(parents=True, exist_ok=True)

        # 2. Load all the checkpoint csv files
        checkpoints = glob.glob(str(dir_cache_pids / '*.csv'))
        checkpoint_start_index = len(checkpoints)
        if checkpoint_start_index > 0:
            dataframes = []
            for checkpoint in checkpoints:
                try:
                    df_checkpoint = pd.read_csv(checkpoint)
                    dataframes.append(df_checkpoint)
                except pd.errors.EmptyDataError:
                    print(f"Warning: {checkpoint} is empty and has been skipped.")
                    continue
            completed_rows = pd.concat(dataframes, ignore_index=True)

            completed_ids = completed_rows['lat_lon_id'].drop_duplicates()

            # Merge on the ID column, keeping track of where each row originates
            merged = df.merge(completed_ids, on='lat_lon_id', how='outer', indicator=True)

            # Filter out rows that come from the 'completed_ids' DataFrame
            df = merged[merged['_merge'] == 'left_only'].drop(columns='_merge')

        def get_street_view_info(longitude, latitude, proxies):
            results = panoids(latitude, longitude, proxies, )
            return results

        def worker(row):
            input_longitude = row.longitude
            input_latitude = row.latitude
            lat_lon_id = row.lat_lon_id
            id_dict = {column: getattr(row, column) for column in id_columns} if id_columns else {}
            return lat_lon_id, (input_longitude, input_latitude), get_street_view_info(input_longitude, input_latitude, self.proxies), id_dict

        # set lat_lon_id if it doesn't exist
        if 'lat_lon_id' not in df.columns:
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
        results = []
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(df) + batch_size - 1) // batch_size
        failed_rows = []
        
        # if there's no rows to process, return completed_ids
        if len(df) == 0:
            return completed_ids
        
        # if not, process the rows
        for i in tqdm(range(num_batches), desc=f"Getting pids by batch size {min(batch_size, len(df))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row): row for row in df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting pids for batch #{i+1}"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['input_latitude'] = input_latitude
                            result['input_longitude'] = input_longitude
                            result['lat_lon_id'] = lat_lon_id
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Error: {e}")
                        failed_rows.append(batch_futures[future])  # Store the failed row

                # Save checkpoint for each batch
                if len(results) > 0:
                    pd.DataFrame(results).to_csv(f'{dir_cache_pids}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
                results = []  # Clear the results list for the next batch

        # Merge all checkpoints into a single dataframe
        results_df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_pids / '*.csv'))], ignore_index=True)

        # Retry failed rows
        if failed_rows:
            print("Retrying failed rows...")
            with ThreadPoolExecutor() as executor:
                retry_futures = {executor.submit(worker, row): row for row in failed_rows}
                for future in tqdm(as_completed(retry_futures), total=len(retry_futures), desc="Retrying failed rows"):
                    try:
                        lat_lon_id, (input_longitude, input_latitude), row_results, id_dict = future.result()
                        for result in row_results:
                            result['input_latitude'] = input_latitude
                            result['input_longitude'] = input_longitude
                            result['lat_lon_id'] = lat_lon_id
                            result.update(id_dict)
                            results.append(result)
                    except Exception as e:
                        print(f"Failed again: {e}")

            # Save the results of retried rows as another checkpoint
            if len(results) > 0:
                pd.DataFrame(results).to_csv(f'{dir_cache_pids}/checkpoint_retry.csv', index=False)
                # Merge the retry checkpoint into the final dataframe
                retry_df = pd.read_csv(f'{dir_cache_pids}/checkpoint_retry.csv')
                results_df = pd.concat([results_df, retry_df], ignore_index=True)

        # now save results_df as a new cache after dropping lat_lon_id
        results_df = results_df.drop(columns='lat_lon_id')
        # drop duplicates in panoid and id_columns
        results_df = results_df.drop_duplicates(subset=['panoid'] + id_columns)
        results_df.to_csv(self.cache_pids_raw, index=False)

        # delete the cache directory
        if dir_cache_pids.exists():
            shutil.rmtree(dir_cache_pids)
        return results_df

    def _get_pids_from_gdf(self, gdf, **kwargs):  
        if self.cache_lat_lon.exists():
            df = pd.read_csv(self.cache_lat_lon)
            print("The lat and lon have been read from the cache")
        else:
            if gdf.crs is None:
                gdf = gdf.set_crs('EPSG:4326')
            elif gdf.crs != 'EPSG:4326':
                # convert to EPSG:4326
                gdf = gdf.to_crs('EPSG:4326')
            # read shapefile
            gp = GeoProcessor(gdf, distance=self.distance, grid=self.grid, grid_size=self.grid_size, **kwargs)
            df = gp.get_lat_lon()
            df['lat_lon_id'] = np.arange(1, len(df) + 1)
            # save df to cache
            df.to_csv(self.cache_lat_lon, index=False)

        if self.cache_pids_raw.exists():
            print("The raw panorama IDs have been read from the cache")
            results_df = pd.read_csv(self.cache_pids_raw)
        else:
            # Use _get_pids_from_df to get pids from df
            results_df = self._get_pids_from_df(df, kwargs["id_columns"])

        # Check if lat and lon are within input polygons
        polygons = gpd.GeoSeries([geom for geom in gdf['geometry'] if geom.type in ['Polygon', 'MultiPolygon']])

        # the rest is only for polygons, so return results_df if there's no polygons
        if len(polygons) == 0:
            return results_df

        # Convert lat, lon to Points and create a GeoSeries
        points = gpd.GeoSeries([Point(lon, lat) for lon, lat in zip(results_df['lon'], results_df['lat'])])

        # Create a GeoDataFrame with the points and an index column
        points_gdf = gpd.GeoDataFrame(geometry=points, crs=gdf.crs)
        points_gdf['index'] = range(len(points_gdf))

        # Create a spatial index on the polygons GeoSeries
        polygons_sindex = polygons.sindex

        # Function to check whether a point is within any polygon
        def is_within_polygon(point):
            possible_matches_index = list(polygons_sindex.intersection(point.bounds))
            possible_matches = polygons.iloc[possible_matches_index]
            precise_matches = possible_matches.contains(point)
            return precise_matches.any()

        # Add progress bar for within_polygon calculation
        with tqdm(total=len(points), desc="Checking points within polygons") as pbar:
            within_polygon = []
            for point in points_gdf['geometry']:
                within_polygon.append(is_within_polygon(point))
                pbar.update()

        results_df['within_polygon'] = within_polygon

        # Return only those points within polygons
        results_within_polygons_df = results_df[results_df['within_polygon']]
        # Drop the 'within_polygon' column
        results_within_polygons_df = results_within_polygons_df.drop(columns='within_polygon')
        return results_within_polygons_df

    def _get_raw_pids(self, **kwargs):
        if self.cache_pids_raw.exists():
            pid = pd.read_csv(self.cache_pids_raw)
            print("The raw panorama IDs have been read from the cache")
            return pid

        if kwargs['lat'] is not None and kwargs['lon'] is not None:
            pid = panoids(kwargs['lat'], kwargs['lon'], self.proxies)
            pid = pd.DataFrame(pid)
            # add input_lat and input_lon
            pid['input_latitude'] = kwargs['lat']
            pid['input_longitude'] = kwargs['lon']
        elif kwargs['input_csv_file'] != "":
            df = pd.read_csv(kwargs['input_csv_file'])
            df = standardize_column_names(df)
            if kwargs['buffer'] > 0:
                gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.longitude, df.latitude), crs='EPSG:4326')
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
                pid = self._get_pids_from_gdf(gdf, **kwargs)
            else:
                pid = self._get_pids_from_df(df, kwargs['id_columns'])
        elif kwargs['input_shp_file'] != "":
            gdf = gpd.read_file(kwargs['input_shp_file'])
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, **kwargs)
        elif kwargs['input_place_name'] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs['input_place_name'])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
            if kwargs['buffer'] > 0:
                gdf = create_buffer_gdf(gdf, kwargs['buffer'])
            pid = self._get_pids_from_gdf(gdf, **kwargs) 
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")

        return pid

    def _get_pids(self, path_pid, **kwargs):
        id_columns = kwargs['id_columns']
        if id_columns is not None:
            if isinstance(id_columns, str):
                id_columns = [id_columns.lower()]
            elif isinstance(id_columns, list):
                id_columns = [column.lower() for column in id_columns]
        else:
            id_columns = []
        # update id_columns
        kwargs['id_columns'] = id_columns
        
        # get raw pid
        pid = self._get_raw_pids(**kwargs)
        
        if kwargs["augment_metadata"] & (self.gsv_api_key != None):
            pid = self._augment_metadata(pid)
        elif kwargs["augment_metadata"] & (self.gsv_api_key == None):
            raise ValueError("Please set the gsv api key by calling the gsv_api_key method.")
        pid.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid)) 

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        # set dir_cache as attribute and create the directory
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        # set other cache directories
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"
        self.cache_pids_augmented = self.dir_cache / "pids_augemented.csv"
        
    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from year and month
        # Fill NA values with a placeholder value
        pid_df['year'] = pid_df['year'].fillna(-1)
        pid_df['month'] = pid_df['month'].fillna(-1)

        # Replace Inf values with a placeholder value
        pid_df['year'] = pid_df['year'].replace([np.inf, -np.inf], -1)
        pid_df['month'] = pid_df['month'].replace([np.inf, -np.inf], -1)

        # Convert to int and then to string, and create the 'date' column
        pid_df['date'] = pid_df['year'].astype(int).astype(str) + "-" + pid_df['month'].astype(int).astype(str)

        # Optionally, you can replace the placeholder values in 'date' column back to NaN
        pid_df['date'] = pid_df['date'].replace('-1--1', np.nan)

        # convert to datetime
        pid_df['date'] = pd.to_datetime(pid_df['date'], format="%Y-%m")
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df['date'] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df['date'] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns='date')
        return pid_df
    
    def download_svi(self, dir_output: str, path_pid: str = None, zoom: int = 2, 
                      cropped: bool = False, full: bool = True, lat: float = None, lon: float = None, 
                      input_csv_file: str = "", input_shp_file: str = "", input_place_name: str = "", 
                      id_columns: Union[str, List[str]] = None, buffer: int = 0, augment_metadata: bool = False, 
                      batch_size: int = 1000, update_pids: bool = False, start_date: str = None, end_date: str = None, 
                      metadata_only: bool = False, **kwargs) -> None:
        """
        Downloads street view images.

        Args:
            dir_output (str): The output directory.
            path_pid (str, optional): The path to the panorama ID file. Defaults to None.
            zoom (int, optional): The zoom level for the images. Defaults to 2.
            cropped (bool, optional): Whether to crop the images. Defaults to False.
            full (bool, optional): Whether to download full images. Defaults to True.
            lat (float, optional): The latitude for the images. Defaults to None.
            lon (float, optional): The longitude for the images. Defaults to None.
            input_csv_file (str, optional): The input CSV file. Defaults to "".
            input_shp_file (str, optional): The input shapefile. Defaults to "".
            input_place_name (str, optional): The input place name. Defaults to "".
            id_columns (Union[str, List[str]], optional): The ID columns. Defaults to None.
            buffer (int, optional): The buffer size. Defaults to 0.
            augment_metadata (bool, optional): Whether to augment the metadata. Defaults to False.
            batch_size (int, optional): The batch size for downloading. Defaults to 1000.
            update_pids (bool, optional): Whether to update the panorama IDs. Defaults to False.
            start_date (str, optional): The start date for the panorama IDs. Format is isoformat (YYYY-MM-DD). Defaults to None.
            end_date (str, optional): The end date for the panorama IDs. Format is isoformat (YYYY-MM-DD). Defaults to None.
            metadata_only (bool, optional): Whether to download metadata only. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            None
        """
        # set necessary directories
        self._set_dirs(dir_output)
        
        # call _get_pids function first if path_pid is None
        if (path_pid is None) & (self.cache_pids_augmented.exists() == False):
            print("Getting pids...")
            path_pid = self.dir_output / "gsv_pids.csv"
            if path_pid.exists() & (update_pids == False):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            id_columns=id_columns, buffer = buffer, augment_metadata=augment_metadata, **kwargs)
        elif self.cache_pids_augmented.exists():
            # copy the cache pids_augmented to path_pid
            path_pid = self.dir_output / "gsv_pids.csv"
            shutil.copy2(self.cache_pids_augmented, path_pid)
            print("The augmented panorama IDs have been saved to {}".format(path_pid))
        # Horizontal Google Street View tiles
        # zoom 3: (8, 4); zoom 5: (26, 13) zoom 2: (4, 2) zoom 1: (2, 1);4:(8,16)
        # zoom = 2
        # h_tiles = 4  # 26
        # v_tiles = 2  # 13
        # cropped = False
        # full = True
        # stop if metadata_only is True
        if metadata_only:
            print("The metadata has been downloaded")
            return
        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "gsv_panorama"
        self.panorama_output.mkdir(parents=True, exist_ok=True)
        
        panoids = self._read_pids(path_pid, start_date, end_date)
        
        if len(panoids) == 0:
            print("There is no panorama ID to download")
            return
        else:
            panoids_rest = self._check_already(panoids)

        if len(panoids_rest) > 0:
            UAs = random.choices(self.user_agents, k = len(panoids_rest))
            # check zoom level is 0<=zoom<=6
            if zoom < 0 or zoom > 6:
                raise ValueError("zoom level should be between 0 and 6")
            h_tiles = 2 ** zoom
            v_tiles = math.ceil(h_tiles / 2)
            ImageTool.dwl_multiple(panoids_rest, zoom, v_tiles, h_tiles, self.panorama_output, UAs, self.proxies, cropped, full, batch_size = batch_size, log_path=self.log_path)
        else:
            print("All images have been downloaded")
        
        # delete the cache directory
        if self.dir_cache.exists():
            shutil.rmtree(self.dir_cache)
            print("The cache directory has been deleted")


# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.api.entities
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the Adapter design for the Entities API of Mapillary API v4.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package Imports
import typing
import json
import ast

# Local imports

# # Utilities
from zensvi.download.mapillary.utils.format import detection_features_to_geojson

# # Models
from zensvi.download.mapillary.models.client import Client

# # Exception Handling
from zensvi.download.mapillary.models.exceptions import InvalidImageKeyError

# # Config
from zensvi.download.mapillary.config.api.entities import Entities

# Library imports
from requests import HTTPError


class EntityAdapter(object):
    """
    Adapter model for dealing with the Entity API, through the DRY principle. The EntityAdapter
    class can be instantiated in the controller modules, providing an abstraction layer that uses
    the Client class, endpoints provided by the API v4 under `/config/api/entities.py`.

    It performs parsing, handling of layers, properties, and fields to make it easier to
    write higher level logic for extracing information, and lets developers to focus only
    on writing the high level business logic without having to repeat the process of parsing
    and using libraries such as `mercantile`, `ast`, and others to only then care about the
    inputs and the outputs

    Usage::

        >>> import mapillary
        >>> from mapillary.models.api.entities import EntityAdapter
        >>> EntityAdapter().fetch_image(image_id='IMAGE_ID', fields=[
        ...     'altitude', 'atomic_scale', 'geometry', 'width'
        ... ])
        >>> EntityAdapter().fetch_map_feature(map_feature_id='MAP_FEATURE_ID', fields=[
        ...         'first_seen_at', 'last_seen_at', 'geometry'
        ...     ])
    """

    def __init__(self):
        """Initializing EntityAdapter constructor"""

        # client object to deal with session and requests
        self.client = Client()

    def fetch_image(
        self, image_id: typing.Union[int, str], fields: list = None
    ) -> dict:
        """
        Fetches images depending on the image_id and the fields provided

        :param image_id: The image_id to extract for
        :type image_id: int

        :param fields: The fields to extract properties for, defaults to []
        :type fields: list

        :return: The fetched GeoJSON
        :rtype: dict
        """

        # Getting the results through the client, and return after decoding
        try:
            return (
                # Converts the dict to a GeoJSON format
                # image_entity_to_geojson(
                # ast converts the str dict to a dict object
                ast.literal_eval(
                    # The results returned here are the str dicts
                    self.client.get(
                        # Calling the endpoint with the parameters ...
                        Entities.get_image(
                            # ... image_id, for the needed image ...
                            image_id=image_id,
                            # ... the fields passed in in ...
                            fields=fields
                            # ... only if the fields are not empty ...
                            if fields != []
                            # ... if they are, get all the fields as a list instead
                            else Entities.get_image_fields(),
                        ),
                        # After retrieval of response, only get the content, decode to utf-8
                    ).content.decode("utf-8")
                )
            )
        except HTTPError:
            # If given ID is an invalid image ID, let the user know
            raise InvalidImageKeyError(image_id)

    def fetch_map_feature(
        self, map_feature_id: typing.Union[int, str], fields: list = None
    ):
        """
        Fetches map features depending on the map_feature_id and the fields provided

        :param map_feature_id: The map_feature_id to extract for
        :type map_feature_id: int

        :param fields: The fields to extract properties for, defaults to []
        :type fields: list

        :return: The fetched GeoJSON
        :rtype: dict
        """

        # Getting the results through the client, and return after decoding
        return ast.literal_eval(
            self.client.get(
                # Calling the endpoint with the parameters ...
                Entities.get_map_feature(
                    # ... image_id, for the needed image ...
                    map_feature_id=map_feature_id,
                    # ... the fields passed in in ...
                    fields=fields
                    # ... only if the fields are not empty ...
                    if fields != []
                    # ... if they are, get all the fields as a list instead
                    else Entities.get_map_feature_fields(),
                ),
                # After retrieval of response, only get the content, decode to utf-8
            ).content.decode("utf-8")
        )

    def fetch_detections(self, identity: int, id_type: bool = True, fields: list = []):
        """
        Fetches detections depending on the id, detections for either map_features or
        images and the fields provided

        :param identity: The id to extract for
        :type identity: int

        :param id_type: Either True(id is for image), or False(id is for map_feature),
            defaults to True
        :type id_type: bool

        :param fields: The fields to extract properties for, defaults to []
        :type fields: list

        :return: The fetched GeoJSON
        :rtype: dict
        """

        # If id_type is True(id is for image)
        if id_type:

            # Store the URL as ...
            url = Entities.get_detection_with_image_id(
                # .. extracted by setting image_id as id ...
                image_id=str(identity),
                # ... passing in the fields given ...
                fields=fields
                # ... only if the fields are not provided empty ...
                if fields != []
                # ... but if they are, set the fields as all possible fields ...
                else Entities.get_detection_with_image_id_fields(),
            )

        # If id_type is False(id is for map_feature)
        else:

            # Store the URL as ...
            url = Entities.get_detection_with_map_feature_id(
                # ... extracted by setting map_feature_id as id ...
                map_feature_id=str(identity),
                # ... passing in the fields given ...
                fields=fields
                # ... only if the fields are not provided empty ...
                if fields != []
                # ... but if they are, set the fields as all possible fields ...
                else Entities.get_detection_with_map_feature_id_fields(),
            )

        # Retrieve the relevant data with `url`, get content, decode to utf-8, return
        return detection_features_to_geojson(
            json.loads(self.client.get(url).content.decode("utf-8"))["data"]
        )

    def is_image_id(self, identity: int, fields: list = None) -> bool:
        """Determines whether the given id is an image_id or a map_feature_id

        :param identity: The ID given to test
        :type identity: int

        :param fields: The fields to extract properties for, defaults to []
        :type fields: list

        :return: True if id is image, else False
        :rtype: bool
        """

        try:
            # If image data is fetched without an exception being thrown ...
            self.fetch_image(image_id=identity, fields=fields)

            # ... return True
            return True

        # An exception to catch the InvalidImageKey exception
        except InvalidImageKeyError:

            # If exception, return False
            return False



import os
import sys
import requests
from PIL import Image
from concurrent.futures import ThreadPoolExecutor, as_completed
import numpy as np
from tqdm import tqdm
import random
from requests.exceptions import ProxyError
import glob

class ImageTool():

    @staticmethod
    def concat_horizontally(im1, im2):
        """
        Description of concat_horizontally
        Horizontally concatenates two images

        Args:
            im1 (undefined): first PIL image
            im2 (undefined): second PIL image

        """
        dst = Image.new('RGB', (im1.width + im2.width, im1.height))
        dst.paste(im1, (0, 0))
        dst.paste(im2, (im1.width, 0))
        return dst

    @staticmethod
    def concat_vertically(im1, im2):
        """
        Description of concat_vertically
        Vertically concatenates two images

        Args:
            im1 (undefined): first PIL image
            im2 (undefined): second PIL image

        """
        dst = Image.new('RGB', (im1.width, im1.height + im2.height))
        dst.paste(im1, (0, 0))
        dst.paste(im2, (0, im1.height))
        return dst
                
    @staticmethod
    def fetch_image_with_proxy(pano_id, zoom, x, y, ua, proxies):
        """
        Fetches an image using a proxy.

        Args:
            pano_id (str): GSV panorama id
            zoom (int): Zoom level for the image
            x (int): The x coordinate of the tile
            y (int): The y coordinate of the tile
            ua (str): User agent string
            proxies (list): A list of available proxies

        Returns:
            Image: The fetched image
        """
        while True:
            # Choose a random proxy for each request
            proxy = random.choice(proxies)
            url_img = f'https://cbk0.google.com/cbk?output=tile&panoid={pano_id}&zoom={zoom}&x={x}&y={y}'
            try:
                image = Image.open(requests.get(url_img, headers=ua, proxies=proxy, stream=True).raw)
                return image
            except ProxyError as e:
                print(f"Proxy {proxy} is not working. Exception: {e}")
                continue

    @staticmethod
    def is_bottom_black(image, row_count=3, intensity_threshold=10):
        """
        Check if the bottom 'row_count' rows of the image are near black, with a given intensity threshold.
        This method uses linear computation instead of nested loops for faster execution.

        Args:
            image (PIL.Image): The image to check.
            row_count (int): Number of rows to check.
            intensity_threshold (int): The maximum intensity for a pixel to be considered black.

        Returns:
            bool: True if the bottom rows are near black, False otherwise.
        """
        # Convert the bottom rows to a numpy array for fast processing
        bottom_rows = np.array(image)[-row_count:, :]
        # Check if all pixels in the bottom rows are less than or equal to the intensity threshold
        return np.all(bottom_rows <= intensity_threshold)

    @staticmethod
    def process_image(image, zoom):
        """
        Crop and resize the image based on zoom level if the bottom is black.

        Args:
            image (PIL.Image): The image to process.
            zoom (int): The zoom level.

        Returns:
            PIL.Image: The processed image.
        """
        if ImageTool.is_bottom_black(image):
            # Compute the crop and resize dimensions based on zoom level
            crop_height, crop_width = 208 * (2 ** zoom), 416 * (2 ** zoom)
            resize_height, resize_width = 256 * (2 ** zoom), 512 * (2 ** zoom)

            # Crop the image
            image = image.crop((0, 0, crop_width, crop_height))

            # Resize the image
            image = image.resize((resize_width, resize_height), Image.ANTIALIAS)

        return image            

    @staticmethod
    def get_and_save_image(pano_id, identif, zoom, vertical_tiles, horizontal_tiles, out_path, ua, proxies, cropped=False, full=True):
        """
        Description of get_and_save_image
        
        Downloads an image tile by tile and composes them together.

        Args:
            pano_id (undefined): GSV anorama id
            identif (undefined): custom identifier
            size (undefined):    image resolution
            vertical_tiles (undefined): number of vertical tiles
            horizontal_tiles (undefined): number of horizontal tiles
            out_path (undefined): output path
            cropped=False (undefined): set True if the image split horizontally in half is needed
            full=True (undefined): set to True if the full image is needed

        """
        for x in range(horizontal_tiles):
            for y in range(vertical_tiles):
                new_img = ImageTool.fetch_image_with_proxy(pano_id, zoom, x, y, ua, proxies)
                if not full:
                    new_img.save(f'{out_path}/{identif}_x{x}_y{y}.jpg')
                if y == 0:
                    first_slice = new_img
                else:
                    first_slice = ImageTool.concat_vertically(first_slice, new_img)

            if x == 0:
                final_image = first_slice
            else:
                final_image = ImageTool.concat_horizontally(final_image, first_slice)

        if full:
            name = f'{out_path}/{identif}'
            if cropped or zoom == 0:
                h_cropped = final_image.size[1] // 2
                final_image = final_image.crop((0, 0, final_image.size[0], h_cropped))

            # Validate image before saving
            if final_image.size[0] > 0 and final_image.size[1] > 0:
                final_image = ImageTool.process_image(final_image, zoom)
                final_image.save(f'{name}.jpg')
            else:
                raise ValueError(f"Invalid image for pano_id {pano_id}")

        return identif


    @staticmethod
    def dwl_multiple(panoids, zoom, v_tiles, h_tiles, out_path, uas, proxies, cropped, full, batch_size = 1000, log_path=None):
        """
        Description of dwl_multiple
        
        Calls the get_and_save_image function using multiple threads.
        
        Args:
            panoids (undefined): GSV anorama id
            zoom (undefined):    image resolution
            v_tiles (undefined): number of vertical tiles
            h_tiles (undefined): number of horizontal tiles
            out_path (undefined): output path
            cropped=False (undefined): set True if the image split horizontally in half is needed
            full=True (undefined): set to True if the full image is needed
            log_path=None (undefined): path to a log file
        """

        if not os.path.exists(out_path):
            os.makedirs(out_path)

        errors = 0

        def log_error(panoid):
            nonlocal log_path
            if log_path is not None:
                with open(log_path, 'a') as log_file:
                    log_file.write(f"{panoid}\n")

        # Calculate current highest batch number
        existing_batches = glob.glob(os.path.join(out_path, "batch_*"))
        existing_batch_numbers = [int(os.path.basename(batch).split('_')[-1]) for batch in existing_batches]
        start_batch_number = max(existing_batch_numbers, default=0)

        num_batches = (len(panoids) + batch_size - 1) // batch_size

        for counter, i in tqdm(enumerate(range(start_batch_number, start_batch_number + num_batches)), desc=f"Downloading images by batch size {min(batch_size, len(panoids))}"):
            # Create a new sub-folder for each batch
            batch_out_path = os.path.join(out_path, f"batch_{i+1}")
            os.makedirs(batch_out_path, exist_ok=True)

            with ThreadPoolExecutor(max_workers=min(len(uas), batch_size)) as executor:
                jobs = []
                batch_panoids = panoids[counter*batch_size : (counter+1)*batch_size]
                batch_uas = uas[counter*batch_size : (counter+1)*batch_size]
                for pano, ua in zip(batch_panoids, batch_uas):
                    kw = {
                        "pano_id": pano,
                        "identif": pano,
                        "ua": ua,
                        "proxies": proxies,
                        "zoom": zoom,
                        "vertical_tiles": v_tiles,
                        "horizontal_tiles": h_tiles,
                        "out_path": batch_out_path,  # Pass the new sub-folder path
                        "cropped": cropped,
                        "full": full
                    }
                    jobs.append(executor.submit(ImageTool.get_and_save_image, **kw))

                for job in tqdm(as_completed(jobs), total=len(jobs), desc=f"Downloading images for batch #{i+1}"):
                    try:
                        job.result()
                    except Exception as e:
                        print(e)
                        errors += 1
                        failed_panoid = batch_panoids[jobs.index(job)]
                        log_error(failed_panoid)

        print("Total images downloaded:", len(panoids) - errors, "Errors:", errors)  


# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.api.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This package contains the class adaptors of the API logic within the Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import entities  # noqa: F401
from . import vector_tiles  # noqa: F401



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.interface
~~~~~~~~~~~~~~~~~~~

This module implements the basic functionalities of the Mapillary Python SDK, a Python
implementation of the Mapillary API v4. For more information, please check out
https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""
# Package level imports
from typing import Union
import requests
import json
import os

# Local
from zensvi.download.mapillary.utils.auth import auth, set_token

# Models
from zensvi.download.mapillary.models.geojson import Coordinates, GeoJSON
from zensvi.download.mapillary.models.config import Config

# Exception classes
from zensvi.download.mapillary.models.exceptions import InvalidOptionError

# Controllers
import zensvi.download.mapillary.controller.image as image
import zensvi.download.mapillary.controller.feature as feature
import zensvi.download.mapillary.controller.detection as detection
import zensvi.download.mapillary.controller.save as save


def configure_mapillary_settings(**kwargs):
    """
    A function allowing the user to configure the Mapillary settings for the session. Takes no
    arguments and sets a global variable used by other functions making API requests. For more
    information what the details of authentication, please check out the blog post at Mapillary.
    https://blog.mapillary.com/update/2021/06/23/getting-started-with-the-new-mapillary-api-v4.html

    :return: None
    :rtype: None

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.configure_mapillary_settings()
        >>> mly.interface.configure_mapillary_settings(use_strict=True)

    :param kwargs: Keyword arguments for the configuration
    :type kwargs: dict

    :param kwargs.use_strict: Whether to use strict mode or not
    :type kwargs.use_strict: bool

    :return: None
    :rtype: None
    """

    return Config(kwargs)


def set_access_token(token: str):
    """
    A function allowing the user to set an access token for the session, which they can create at
    https://www.mapillary.com/dashboard/developers. Takes token as an argument and sets a global
    variable used by other functions making API requests. For more information what the details
    of authentication, please check out the blog post at Mapillary.
    https://blog.mapillary.com/update/2021/06/23/getting-started-with-the-new-mapillary-api-v4.html

    :param token: The token itself that would
        be set and accessed globally. Must be obtained
    :type token: str

    :return: None
    :rtype: None

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
    """

    return set_token(token)


@auth()
def get_image_close_to(latitude=-122.1504711, longitude=37.485073, **kwargs):
    """
    Function that takes a longitude, latitude as argument and outputs the near images. This
    makes an API call with the token set in set_access_token and returns a JSON object.

    :param longitude: The longitude
    :type longitude: float or double

    :param latitude: The latitude
    :type latitude: float or double

    :param kwargs.fields: A list of options, either as ['all'], or a list of fields.
        See https://www.mapillary.com/developer/api-documentation/, under 'Fields' for more insight.
    :type kwargs.fields: list

    :param kwargs.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type kwargs.zoom: int

    :param kwargs.radius: The radius of the images obtained from a center center
    :type kwargs.radius: float or int or double

    :param kwargs.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'both'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type kwargs.image_type: str

    :param kwargs.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type kwargs.min_captured_at: str

    :param kwargs.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type kwargs.max_captured_at: str

    :param kwargs.org_id: The organization id, ID of the organization this image (or sets of
        images) belong to. It can be absent. Thus, default is -1 (None)
    :type kwargs.org_id: int

    :return: GeoJSON
    :rtype: dict

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
        >>> mly.interface.get_image_close_to(longitude=31, latitude=30)
        ... {'type': 'FeatureCollection', 'features': [{'type': 'Feature',
        'geometry': {'type': 'Point', 'coordinates': [30.9912246465683,
        29.99794091267283]}, 'properties': {'captured_at': 1621008070596,
        'compass_angle': 322.56726074219, 'id': 499412381300321, 'is_pano':
        False, 'sequence_id': '94afmyyhq85xd9bi8p44ve'}} ...
    """

    return image.get_image_close_to_controller(
        latitude=latitude,
        longitude=longitude,
        kwargs=kwargs,
    )


@auth()
def get_image_looking_at(
    at: dict,
    **filters: dict,
) -> GeoJSON:
    """
    Function that takes two sets of latitude and longitude, where the 2nd set is the
    "looking at" location from 1st set's perspective argument and outputs the near images. This
    makes an API call with the token set in set_access_token and returns a JSON object.

    :param at: The coordinate sets to where a certain point is being looked at

        Format::

            >>> {
            ...     'lng': 'longitude',
            ...     'lat': 'latitude'
            ... }

    :type at: dict

    :param filters.min_captured_at: The minimum date to filter till
    :type filters.min_captured_at: str

    :param filters.max_captured_at: The maximum date to filter upto
    :type filters.max_captured_at: str

    :param filters.radius: The radius that the geometry points will lie in
    :type filters.radius: float

    :param filters.image_type: Either 'pano', 'flat' or 'all'
    :type filters.image_type: str

    :param filters.organization_id: The organization to retrieve the data for
    :type filters.organization_id: str

    :return: The GeoJSON response containing relevant features
    :rtype: GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.get_image_looking_at(
        ...         at={
        ...             'lng': 12.955075073889,
        ...             'lat': 48.053805939722,
        ...         },
        ...         radius = 5000,
        ...     )
        >>> data
        ... {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry': {'type':
        ... 'Point', 'coordinates': [12.954479455947876, 48.05091893670834]}, 'properties':
        ... {'captured_at': 1612606959408, 'compass_angle': 21.201110839844, 'id': 1199705400459580,
        ... 'is_pano': False, 'sequence_id': 'qrrqtke4a6vtygyc7w8rzc'}}, ... }
    """

    return image.get_image_looking_at_controller(
        at=at,
        filters=filters,
    )


@auth()
def is_image_being_looked_at(
    at: Union[dict, Coordinates, list],
    **filters: dict,
) -> bool:
    """
    Function that two sets of coordinates and returns whether the image  with coordinates of "at"
    is looked at or not by the image with coordinates of "looker".

    :param at: The coordinate sets to where a certain point is being looked at

        Format::

            >>> at_dict = {
            ...     'lng': 'longitude',
            ...     'lat': 'latitude'
            ... }
            >>> at_list = [12.954940544167, 48.0537894275]
            >>> from mapillary.models.geojson import Coordinates
            >>> at_coord: Coordinates = Coordinates(lng=12.954940544167, lat=48.0537894275)

    :type at: Union[dict, mapillary.models.geojson.Coordinates, list]

    :return: True if the image is looked at, False otherwise
    :rtype: bool

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.is_image_being_looked_at(
        ...         at={
        ...             'lng': 12.955075073889,
        ...             'lat': 48.053805939722,
        ...         },
        ...         radius=50,
        ...     )
        ... True
        >>> # OR
        >>> from mapillary.models.geojson import Coordinates
        >>> mly.interface.is_image_looked_at(
        ...         at=Coordinates(lng=11.954940544167, lat=46.0537894275),
        ...         radius=50,
        ...     )
        ... True
    """

    return image.is_image_being_looked_at_controller(at=at, filters=filters)


@auth()
def get_detections_with_image_id(image_id: int, fields: list = []):
    """
    Extracting all the detections within an image using an image key

    :param image_id: The image key as the argument
    :type image_id: int

    :param fields: The fields possible for the detection endpoint. Please see
        https://www.mapillary.com/developer/api-documentation for more information
    :type fields: list

    :return: The GeoJSON in response
    :rtype: dict

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
        >>> mly.interface.get_detections_with_image_id(image_id=1933525276802129)
        ... {"data":[{"created_at":"2021-05-20T17:49:01+0000","geometry":
        ... "GjUKBm1weS1vchIVEgIAABgDIg0JhiekKBoqAABKKQAPGgR0eXBlIgkKB3BvbHlnb24ogCB4AQ==","image"
        ... :{"geometry":{"type":"Point","coordinates":[-97.743279722222,30.270651388889]},"id":
        ... "1933525276802129"},"value":"regulatory--no-parking--g2","id":"1942105415944115"},
        ... {"created_at":"2021-05-20T18:40:21+0000","geometry":
        ... "GjYKBm1weS1vchIWEgIAABgDIg4J7DjqHxpWAADiAVUADxoEdHlwZSIJCgdwb2x5Z29uKIAgeAE=",
        ... "image":{"geometry":{"type":"Point","coordinates":[-97.743279722222,30.270651388889]},
        ... "id":"1933525276802129"},"value":"information--parking--g1","id":"1942144785940178"},
        ... , ...}
    """

    return detection.get_image_detections_controller(
        image_id=image_id,
        fields=fields,
    )


@auth()
def get_detections_with_map_feature_id(
    map_feature_id: str, fields: list = None
) -> GeoJSON:
    """
    Extracting all detections made for a map feature key

    :param map_feature_id: A map feature key as the argument
    :type map_feature_id: int

    :param fields: The fields possible for the detection endpoint. Please see
        https://www.mapillary.com/developer/api-documentation for more information
    :type fields: list

    :return: The GeoJSON in response
    :rtype: GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.get_detections_with_map_feature_id(map_feature_id='1933525276802129')
        ...     File "/home/saif/MLH/mapillary-python-sdk/mapillary/controller/rules/verify.py",
        ...         line 227, in valid_id
        ...             raise InvalidOptionError(
        ... mly.models.exceptions.InvalidOptionError: InvalidOptionError: Given id value,
        ...     "Id: 1933525276802129, image: False" while possible id options, [Id is image_id
        ...     AND image is True, key is map_feature_id ANDimage is False]
    """

    return detection.get_map_feature_detections_controller(
        map_feature_id=map_feature_id,
        fields=fields,
    )


@auth()
def image_thumbnail(image_id: str, resolution: int = 1024) -> str:
    """
    Gets the thumbnails of images from the API

    :param image_id: Image key as the argument

    :param resolution: Option for the thumbnail size, with available resolutions:
        256, 1024, and 2048

    :return: A URL for the thumbnail
    :rtype: str

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.image_thumbnail(
        ...     image_id='IMAGE_ID_HERE', resolution=1024
        ... )
    """

    return image.get_image_thumbnail_controller(image_id, resolution=resolution)


@auth()
def images_in_bbox(bbox: dict, **filters) -> str:
    """
    Gets a complete list of images with custom filter within a BBox

    :param bbox: Bounding box coordinates

        Format::

            >>> {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bbox: dict

    :param filters: Different filters that may be applied to the output

        Example filters::

            - max_captured_at
            - min_captured_at
            - image_type: pano, flat, or all
            - compass_angle
            - sequence_id
            - organization_id

    :type filters: dict

    :return: Output is a GeoJSON string that represents all the within a bbox after passing given
        filters
    :rtype: str

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.images_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     max_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     min_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     image_type='pano',
        ...     compass_angle=(0, 360),
        ...     sequence_id='SEQUENCE_ID',
        ...     organization_id='ORG_ID'
        ... )
    """

    return image.get_images_in_bbox_controller(
        bounding_box=bbox, layer="image", zoom=14, filters=filters
    )


@auth()
def sequences_in_bbox(bbox: dict, **filters) -> str:
    """
    Gets a complete list of all sequences of images that satisfy given filters
    within a BBox.

    :param bbox: Bounding box coordinates

        Example::

            >>> _ = {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bbox: dict

    :param filters: Different filters that may be applied to the output

        Example filters::

            - max_captured_at
            - min_captured_at
            - image_type: pano, flat, or all
            - org_id

    :type filters: dict

    :return: Output is a GeoJSON string that contains all the filtered sequences within a bbox.
        Sequences would NOT be cut at BBox boundary, would select all sequences which are partially
        or entirely in BBox
    :rtype: str

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.sequences_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     max_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     min_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     image_type='pano',
        ...     org_id='ORG_ID'
        ... )
    """

    return image.get_images_in_bbox_controller(
        bounding_box=bbox, layer="sequence", zoom=14, filters=filters
    )


@auth()
def map_feature_points_in_bbox(
    bbox: dict, filter_values: list = None, **filters: dict
) -> str:
    """
    Extracts map feature points within a bounding box (bbox)

    :param bbox: bbox coordinates as the argument

        Example::

            >>> _ = {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bbox: dict

    :param filter_values: a list of filter values supported by the API

        Example::

            >>> _ = ['object--support--utility-pole', 'object--street-light']

    :type filter_values: list

    :param filters: kwarg filters to be applied on the resulted GeoJSON

        Chronological filters,

        - *existed_at*: checks if a feature existed after a certain date depending on the time
            it was first seen at.
        - *existed_before*: filters out the features that existed after a given date

    :type filters: dict

    :return: GeoJSON Object
    :rtype: dict

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.map_feature_points_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     filter_values=['object--support--utility-pole', 'object--street-light'],
        ...     existed_at='YYYY-MM-DD HH:MM:SS',
        ...     existed_before='YYYY-MM-DD HH:MM:SS'
        ... )
    """

    return feature.get_map_features_in_bbox_controller(
        bbox=bbox, filters=filters, filter_values=filter_values, layer="points"
    )


@auth()
def traffic_signs_in_bbox(
    bbox: dict, filter_values: list = None, **filters: dict
) -> str:
    """
    Extracts traffic signs within a bounding box (bbox)

    :param bbox: bbox coordinates as the argument

        Example::

            >>> {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bbox: dict

    :param filter_values: a list of filter values supported by the API,

        Example::

            >>> ['regulatory--advisory-maximum-speed-limit--g1', 'regulatory--atvs-permitted--g1']

    :type filter_values: list

    :param filters: kwarg filters to be applied on the resulted GeoJSON

        Chronological filters,

        - *existed_at*: checks if a feature existed after a certain date depending on the time
            it was first seen at.
        - *existed_before*: filters out the features that existed after a given date

    :type filters: dict

    :return: GeoJSON Object
    :rtype: dict

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.traffic_signs_in_bbox(
        ...    bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...    },
        ...    filter_values=[
        ...        'regulatory--advisory-maximum-speed-limit--g1',
        ...        'regulatory--atvs-permitted--g1'
        ...    ],
        ...    existed_at='YYYY-MM-DD HH:MM:SS',
        ...    existed_before='YYYY-MM-DD HH:MM:SS'
        ... )
    """

    return feature.get_map_features_in_bbox_controller(
        bbox=bbox, filters=filters, filter_values=filter_values, layer="traffic_signs"
    )


@auth()
def images_in_geojson(geojson: dict, dir_cache: str = None, **filters: dict):
    """
    Extracts all images within a shape

    :param geojson: A geojson as the shape acting as the query extent
    :type geojson: dict

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :return: A GeoJSON object
    :rtype: mapillary.models.geojson.GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> from mapillary.models.geojson import GeoJSON
        >>> import json
        >>> mly.interface.set_access_token('MLY|YYY')
        >>> data = mly.interface.images_in_geojson(json.load(open('my_geojson.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """

    return image.geojson_features_controller(
        geojson=geojson, is_image=True, filters=filters, dir_cache = dir_cache
    )


@auth()
def images_in_shape(shape, **filters: dict):
    """
    Extracts all images within a shape or polygon.

    Format::

        >>> {
        ...    "type": "FeatureCollection",
        ...     "features": [
        ...        {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                         [
        ...                             7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

    :param shape: A shape that describes features, formatted as a geojson
    :type shape: dict

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :return: A GeoJSON object
    :rtype: mapillary.models.geojson.GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.images_in_shape(json.load(open('polygon.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """

    return image.shape_features_controller(shape=shape, is_image=True, filters=filters)


@auth()
def map_features_in_geojson(geojson: dict, **filters: dict):
    """
    Extracts all map features within a geojson's boundaries

    :param geojson: A geojson as the shape acting as the query extent
    :type geojson: dict

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type filters.zoom: int

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :return: A GeoJSON object
    :rtype: mapillary.models.geojson.GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|YYY')
        >>> data = mly.interface.map_features_in_geojson(
        ...     json.load(
        ...         open('my_geojson.geojson', mode='r')
        ...     )
        ... )
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """

    if isinstance(geojson, str):
        if "http" in geojson:
            geojson = json.loads(requests.get(geojson).content.decode("utf-8"))

    return image.geojson_features_controller(
        geojson=geojson, is_image=False, filters=filters, 
    )


@auth()
def map_features_in_shape(shape: dict, **filters: dict):
    """
    Extracts all map features within a shape/polygon

    Format::

        >>> _ = {
        ...     "type": "FeatureCollection",
        ...     "features": [
        ...         {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                         [
        ...                             7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

    :param shape: A shape that describes features, formatted as a geojson
    :type shape: dict

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type filters.zoom: int

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :return: A GeoJSON object
    :rtype: mapillary.models.geojson.GeoJSON

    Usage::

        >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.map_features_in_shape(json.load(open('polygon.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """

    if isinstance(shape, str):
        if "http" in shape:
            shape = json.loads(requests.get(shape).content.decode("utf-8"))

    return image.shape_features_controller(shape=shape, is_image=False, filters=filters)


@auth()
def feature_from_key(key: str, fields: list = []) -> str:
    """
    Gets a map feature for the given key argument

    :param key: The map feature ID to which will be used to get the feature
    :type key: int

    :param fields: The fields to include. The field 'geometry' will always be included
        so you do not need to specify it, or if you leave it off, it will still be returned.

        Fields::

                1. first_seen_at - timestamp, timestamp of the least recent
                    detection contributing to this feature
                2. last_seen_at - timestamp, timestamp of the most recent
                    detection contributing to this feature
                3. object_value - string, what kind of map feature it is
                4. object_type - string, either a traffic_sign or point
                5. geometry - GeoJSON Point geometry
                6. images - list of IDs, which images this map feature was derived
                from

        Refer to https://www.mapillary.com/developer/api-documentation/#map-feature for more details

    :type fields: list

    :return: A GeoJSON string that represents the queried feature
    :rtype: str

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.feature_from_key(
        ...     key='VALID_MAP_FEATURE_KEY',
        ...     fields=['object_value']
        ... )
    """

    return feature.get_feature_from_key_controller(key=int(key), fields=fields)


@auth()
def image_from_key(key: str, fields: list = None) -> str:
    """
    Gets an image for the given key argument

    :param key: The image unique key which will be used for image retrieval
    :type key: int

    :param fields: The fields to include. The field 'geometry' will always be included
        so you do not need to specify it, or if you leave it off, it will still be returned.

        Fields,

        1. altitude - float, original altitude from Exif
        2. atomic_scale - float, scale of the SfM reconstruction around the image
        3. camera_parameters - array of float, intrinsic camera parameters
        4. camera_type - enum, type of camera projection (perspective, fisheye, or
            spherical)
        5. captured_at - timestamp, capture time
        6. compass_angle - float, original compass angle of the image
        7. computed_altitude - float, altitude after running image processing
        8. computed_compass_angle - float, compass angle after running image processing
        9. computed_geometry - GeoJSON Point, location after running image processing
        10. computed_rotation - enum, corrected orientation of the image
        11. exif_orientation - enum, orientation of the camera as given by the exif tag
            (see: https://sylvana.net/jpegcrop/exif_orientation.html)
        12. geometry - GeoJSON Point geometry
        13. height - int, height of the original image uploaded
        14. thumb_256_url - string, URL to the 256px wide thumbnail
        15. thumb_1024_url - string, URL to the 1024px wide thumbnail
        16. thumb_2048_url - string, URL to the 2048px wide thumbnail
        17. merge_cc - int, id of the connected component of images that were aligned
            together
        18. mesh - { id: string, url: string } - URL to the mesh
        19. quality_score - float, how good the image is (experimental)
        20. sequence - string, ID of the sequence
        21. sfm_cluster - { id: string, url: string } - URL to the point cloud
        22. width - int, width of the original image uploaded

        Refer to https://www.mapillary.com/developer/api-documentation/#image for more details

    :type fields: list

    :return: A GeoJSON string that represents the queried image
    :rtype: str

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.image_from_key(
        ...     key='VALID_IMAGE_KEY',
        ...     fields=['captured_at', 'sfm_cluster', 'width']
        ... )
    """

    return image.get_image_from_key_controller(key=int(key), fields=fields)


@auth()
def save_locally(
    geojson_data: str,
    file_path: str = os.path.dirname(os.path.realpath(__file__)),
    file_name: str = None,
    extension: str = "geojson",
) -> None:
    """
    This function saves the geojson data locally as a file
    with the given file name, path, and format.

    :param geojson_data: The GeoJSON data to be stored
    :type geojson_data: str

    :param file_path: The path to save the data to. Defaults to the current directory path
    :type file_path: str

    :param file_name: The name of the file to be saved. Defaults to 'geojson'
    :type file_name: str

    :param extension: The format to save the data as. Defaults to 'geojson'
    :type extension: str

    Note::

        Allowed file format values at the moment are,
            - geojson
            - CSV

    *TODO*: More file format will be supported further in developemtn
    *TODO*: Suggestions and help needed at mapillary/mapillary-python-sdk!

    :return: None
    :rtype: None

    Usage::

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.save_locally(
        ...     geojson_data=geojson_data,
        ...     file_path=os.path.dirname(os.path.realpath(__file__)),
        ...     file_name='test_geojson',
        ...     extension='geojson'
        ... )
        >>> mly.interface.save_locally(
        ...     geojson_data=geojson_data,
        ...     file_path=os.path.dirname(os.path.realpath(__file__)),
        ...     file_name='local_geometries',
        ...     extension='csv'
        ... )
    """

    # Check if a valid file format was provided
    if extension.lower() not in ["geojson", "csv"]:
        # If not, raise an error
        raise InvalidOptionError(
            param="format",
            value=extension,
            options=["geojson", "csv"],
        )

    return (
        save.save_as_geojson_controller(
            data=geojson_data, path=file_path, file_name=file_name
        )
        if extension.lower() == "geojson"
        else save.save_as_csv_controller(
            data=geojson_data, path=file_path, file_name=file_name
        )
    )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controllers.detection
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the detection based business logic functionalities of the Mapillary
Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import typing

# Local imports

# # Adapter Imports
from zensvi.download.mapillary.models.api.entities import EntityAdapter

# # Models
from zensvi.download.mapillary.models.geojson import GeoJSON

# # Rules
from zensvi.download.mapillary.utils.verify import valid_id


def get_image_detections_controller(
    image_id: typing.Union[str, int], fields: list = []
) -> GeoJSON:
    """
    Get image detections with given (image) key

    :param image_id: The image id
    :type image_id: str

    :param fields: The fields possible for the detection endpoint. Please see
        https://www.mapillary.com/developer/api-documentation for more information
    :type fields: list

    :return: GeoJSON
    :rtype: dict
    """

    # Checks if the Id given is indeed a valid image_id
    valid_id(identity=image_id, image=True)

    # Return results from the Adapter
    return GeoJSON(
        geojson=EntityAdapter().fetch_detections(
            identity=image_id,
            id_type=True,
            fields=fields,
        )
    )


def get_map_feature_detections_controller(
    map_feature_id: typing.Union[str, int], fields: list = []
) -> GeoJSON:
    """
    Get image detections with given (map feature) key

    :param map_feature_id: The map feature id
    :type map_feature_id: str

    :param fields: The fields possible for the detection endpoint. Please see
        https://www.mapillary.com/developer/api-documentation for more information
    :type fields: list

    :return: GeoJSON
    :rtype: dict
    """

    # Checks if the Id given is indeed a valid image_id
    valid_id(identity=map_feature_id, image=False)

    # Return results from the Adapter
    return GeoJSON(
        geojson=EntityAdapter().fetch_detections(
            identity=map_feature_id,
            id_type=False,
            fields=fields,
        )
    )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.__init__
~~~~~~~~~~~~~~~~~~

This module imports the necessary parts of the SDK

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Class models for representing the classes
from . import models  # noqa: F401

# Utilities for business logic
from . import utils  # noqa: F401

# Configurations for the API
from . import config  # noqa: F401

# Business logic for the end API calls
from . import controller  # noqa: F401

# The main interface
from zensvi.download.mapillary import interface  # noqa: F401



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.time
====================

This module contains the time utilies for the UNIX epoch seconds, the time and the date range, and
the date filtering logic.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import datetime
import re


def date_to_unix_timestamp(date: str) -> int:
    """
    A utility function that converts the given date
    into its UNIX epoch timestamp equivalent. It accepts the formats, ranging from
    YYYY-MM-DDTHH:MM:SS, to simply YYYY, and a permutation of the fields in between as well

    Has a special argument, '*', which returns current timestamp

    :param date: The date to get the UNIX timestamp epoch of
    :type date: str

    :return: The UNIX timestamp equivalent of the input date
    :rtype: int

    Usage::

        >>> from utils.time_utils import date_to_unix_timestamp
        >>> date_to_unix_timestamp('2020-10-23')
        ... "1603393200"
    """

    # Returns the epoch current timestamp in milliseconds
    if date == "*":
        return int(datetime.datetime.now().timestamp()) * 1000

    # Return the epoch timestamp in miliseconds
    return int(datetime.datetime.fromisoformat(date).timestamp()) * 1000


def is_iso8601_datetime_format(date_time: str) -> bool:
    """
    Checks if the date time is in ISO 8601 format

    :param date_time: The date time to be checked
    :type date_time: str

    :return: True if the date time is in ISO 8601 format, else False
    :rtype: bool
    """

    return (
        re.match(r"(\d{4})\-(\d{2})\-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})Z", date_time)
        is not None
    )



import random
import datetime
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, as_completed
import warnings
import requests
from pathlib import Path
import geopandas as gpd
from tqdm import tqdm
from shapely.geometry import Point
import warnings
from shapely.errors import ShapelyDeprecationWarning
import json
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)

import glob
import shutil
import numpy as np
import osmnx as ox
from PIL import Image

from zensvi.download.base import BaseDownloader
import zensvi.download.mapillary.interface as mly
from zensvi.download.utils.geoprocess import GeoProcessor
from zensvi.download.utils.helpers import standardize_column_names, create_buffer_gdf, check_and_buffer

# set logging level to warning
import logging
logging.getLogger('mapillary.utils.client').setLevel(logging.WARNING)

class MLYDownloader(BaseDownloader):
    def __init__(self, mly_api_key, log_path = None, distance = 1, grid = False, grid_size = 1, max_workers = None):
        super().__init__(log_path, distance, grid, grid_size) 
        self._mly_api_key = mly_api_key
        self._max_workers = max_workers
        mly.set_access_token(self.mly_api_key)
        
    @property
    def mly_api_key(self):
        return self._mly_api_key
    @mly_api_key.setter
    def mly_api_key(self,mly_api_key):
        self._mly_api_key = mly_api_key
    
    @property
    def max_workers(self):
        return self._max_workers
    @max_workers.setter
    def max_workers(self,max_workers):
        self._max_workers = max_workers
    
    def _read_pids(self, path_pid):
        pid_df = pd.read_csv(path_pid)
        # drop NA values in id columns
        pid_df = pid_df.dropna(subset=['id'])
        # get unique pids (ie "id" columns) as a list
        pids = pid_df["id"].astype('int64').unique().tolist()
        return pids

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        self.pids_url = self.dir_output / "pids_urls.csv"
        # set dir_cache as attribute and create the directory
        self.dir_cache = self.dir_output / "cache_zensvi"
        self.dir_cache.mkdir(parents=True, exist_ok=True)
        # set other cache directories
        self.cache_lat_lon = self.dir_cache / "lat_lon.csv"
        self.cache_pids_raw = self.dir_cache / "pids_raw.csv"
    
    def _get_pids_from_gdf(self, gdf, mly_kwargs, **kwargs):
        # set crs to EPSG:4326 if it's None
        if gdf.crs is None:
            gdf = gdf.set_crs('EPSG:4326')
        elif gdf.crs != 'EPSG:4326':
            # convert to EPSG:4326
            gdf = gdf.to_crs('EPSG:4326')
        
        # convert to FeatureCollection
        geojson = json.loads(gdf.to_json())
        
        # use get pids with mly.interface.images_in_geojson
        result_json = mly.images_in_geojson(geojson, dir_cache = self.dir_cache, **mly_kwargs).to_dict()["features"]

        # convert to geodataframe
        result_gdf = gpd.GeoDataFrame.from_features(result_json)
        
        # add lon and lat columns
        result_gdf['lon'] = result_gdf['geometry'].apply(lambda geom: geom.x)
        result_gdf['lat'] = result_gdf['geometry'].apply(lambda geom: geom.y)
        
        # drop geometry column
        result_df = result_gdf.drop(columns='geometry')
        
        return result_df
        
    def _get_raw_pids(self, **kwargs):
        mly_allowed_keys = {'fields', 'zoom', 'radius', 'image_type', 'min_captured_at', 'max_captured_at', 'organization_id'} 
        mly_kwargs = {k: v for k, v in kwargs.items() if k in mly_allowed_keys}
        if self.cache_pids_raw.exists():
            pid = pd.read_csv(self.cache_pids_raw)
            print("The raw panorama IDs have been read from the cache")
            return pid
        
        # input: lat and lon
        if kwargs['lat'] is not None and kwargs['lon'] is not None:
            # create a geodataframe with lat and lon
            df = pd.DataFrame({'lat': [kwargs['lat']], 'lon': [kwargs['lon']]})
            gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.lon, df.lat), crs='EPSG:4326')
        
        # input: csv file
        elif kwargs['input_csv_file'] != "":
            df = pd.read_csv(kwargs['input_csv_file'])
            df = standardize_column_names(df)
            gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.longitude, df.latitude), crs='EPSG:4326')

        # input: shapefile
        elif kwargs['input_shp_file'] != "":
            gdf = gpd.read_file(kwargs['input_shp_file'])

        # input: place name
        elif kwargs['input_place_name'] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs['input_place_name'])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")
        
        # check geometry type and buffer
        gdf = check_and_buffer(gdf, kwargs['buffer'])
        # get pid
        pid = self._get_pids_from_gdf(gdf, mly_kwargs, **kwargs)

        return pid
    
    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from captured_at (milliseconds from Unix epoch)
        pid_df['date'] = pd.to_datetime(pid_df['captured_at'], unit='ms')
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df['date'] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df['date'] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns='date')
        return pid_df

    def _get_pids(self, path_pid, **kwargs):
        # get raw pid
        pid = self._get_raw_pids(**kwargs)
        
        if pid is None:
            print("There is no panorama ID to download")
            return

        # # Assuming that df is your DataFrame and that 'geometry' column contains Point objects
        # # convert the geometry column into shapely geometry objects
        # pid['geometry'] = pid['geometry'].apply(wkt.loads)
        # pid['lon'] = pid['geometry'].apply(lambda geom: geom.x)
        # pid['lat'] = pid['geometry'].apply(lambda geom: geom.y)
        # # drop geometry column
        # pid = pid.drop(columns='geometry')
        # move the "id" column to the first column
        pid = pid[["id"] + [col for col in pid.columns if col != "id"]]

        # keep id,captured_at,compass_angle,is_pano,organization_id,sequence_id,input_latitude,input_longitude,lon,lat drop other columns
        pid = pid[['id','captured_at','compass_angle','is_pano','organization_id','sequence_id','lon','lat']]

        pid.to_csv(path_pid, index=False)
        print("The panorama IDs have been saved to {}".format(path_pid)) 
    
    def _get_urls_mly(self, path_pid, resolution=1024):
        # check if seld.cache_pids_urls exists
        if self.pids_url.exists():
            print("The panorama URLs have been read from the cache")
            return
        
        dir_cache_urls = self.dir_cache / 'urls'
        dir_cache_urls.mkdir(parents=True, exist_ok=True)

        checkpoints = glob.glob(str(dir_cache_urls / '*.csv'))
        checkpoint_start_index = len(checkpoints)

        panoids = set(self._read_pids(path_pid))  # Convert to set for faster operations
        if len(panoids) == 0:
            print("There is no panorama ID to download")
            return

        # Read all panoids from the checkpoint files
        completed_panoids = set()  # Use set for faster operations
        for checkpoint in checkpoints:
            try:
                df_checkpoint = pd.read_csv(checkpoint)
                completed_panoids.update(df_checkpoint['id'].tolist())
            except pd.errors.EmptyDataError:
                print(f"Warning: {checkpoint} is empty and has been skipped.")
                continue

        # Filter out the panoids that have already been processed
        panoids = list(panoids - completed_panoids)  # Subtract sets and convert back to list
        if len(panoids) == 0:
            print("All images have been downloaded")
            return

        def worker(panoid, resolution):
            url = mly.image_thumbnail(panoid, resolution=resolution)
            return panoid, url

        results = {}
        batch_size = 1000  # Modify this to a suitable value
        num_batches = (len(panoids) + batch_size - 1) // batch_size

        for i in tqdm(range(num_batches), desc=f"Getting urls by batch size {min(batch_size, len(panoids))}"):
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, panoid, resolution): panoid for panoid in panoids[i*batch_size : (i+1)*batch_size]}
                
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Getting urls for batch #{i+1}"):
                    current_panoid = batch_futures[future]
                    try:
                        panoid, url = future.result()
                        results[panoid] = url
                    except Exception as e:
                        print(f"Error: {e}")
                        self._log_write(current_panoid)
                        continue

            if len(results) > 0:
                pd.DataFrame.from_dict(results, orient='index').reset_index().rename(columns={'index': 'id', 0: 'url'}).to_csv(f'{dir_cache_urls}/checkpoint_batch_{checkpoint_start_index+i+1}.csv', index=False)
            results = {}

        # Merge all checkpoints into a single dataframe
        results_df = pd.concat([pd.read_csv(checkpoint) for checkpoint in glob.glob(str(dir_cache_urls / '*.csv'))], ignore_index=True)
        results_df.to_csv(self.pids_url, index=False)

        if dir_cache_urls.exists():
            shutil.rmtree(dir_cache_urls)


    def _download_images_mly(self, path_pid, cropped, batch_size, start_date, end_date):
        checkpoints = glob.glob(str(self.panorama_output / '**/*.png'), recursive=True)

        # Read already downloaded images and convert to ids
        downloaded_ids = set([Path(file_path).stem for file_path in checkpoints])  # Use set for faster operations

        pid_df = pd.read_csv(path_pid).dropna(subset=['id'])
        pid_df["id"] = pid_df["id"].astype('int64')
        urls_df = pd.read_csv(self.pids_url)
        urls_df["id"] = urls_df["id"].astype('int64')
        # merge pid_df and urls_df
        urls_df = urls_df.merge(pid_df, on='id', how='left')
        # filter out the rows by date
        urls_df = self._filter_pids_date(urls_df, start_date, end_date)

        # Filter out the ids that have already been processed
        urls_df = urls_df[~urls_df['id'].isin(downloaded_ids)]  # Use isin for efficient operation

        def worker(row, output_dir, cropped):
            url, panoid = row.url, row.id
            user_agent = random.choice(self.user_agents)
            proxy = random.choice(self.proxies)

            image_name = f'{panoid}.png'  # Use id for file name
            image_path = output_dir / image_name
            try:
                response = requests.get(url, headers=user_agent, proxies=proxy, timeout=10)
                if response.status_code == 200:
                    with open(image_path, 'wb') as f:
                        f.write(response.content)
                    
                    if cropped:
                        img = Image.open(image_path)
                        w, h = img.size
                        img_cropped = img.crop((0, 0, w, h // 2))
                        img_cropped.save(image_path)

                else:
                    self._log_write(panoid)
            except Exception as e:
                self._log_write(panoid)
                print(f"Error: {e}" )

        num_batches = (len(urls_df) + batch_size - 1) // batch_size

        # Calculate current highest batch number
        existing_batches = glob.glob(str(self.panorama_output / "batch_*"))
        existing_batch_numbers = [int(Path(batch).name.split('_')[-1]) for batch in existing_batches]
        start_batch_number = max(existing_batch_numbers, default=0)

        for i in tqdm(range(start_batch_number, start_batch_number + num_batches), desc=f"Downloading images by batch size {min(batch_size, len(urls_df))}"):
            # Create a new sub-folder for each batch
            batch_out_path = self.panorama_output / f"batch_{i+1}"
            batch_out_path.mkdir(exist_ok=True)
            
            with ThreadPoolExecutor() as executor:
                batch_futures = {executor.submit(worker, row, batch_out_path, cropped): row.id for row in urls_df.iloc[i*batch_size : (i+1)*batch_size].itertuples()}
                for future in tqdm(as_completed(batch_futures), total=len(batch_futures), desc=f"Downloading images for batch #{i+1}"):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"Error: {e}")

                
    def download_svi(self, dir_output, path_pid = None, lat=None, lon=None, input_csv_file="", input_shp_file = "", input_place_name =
                    "", buffer = 0, update_pids = False, resolution = 1024, cropped = False, batch_size = 1000,
                    start_date = None, end_date = None, metadata_only = False, **kwargs):
        # set necessary directories
        self._set_dirs(dir_output)
        
        # call _get_pids function first if path_pid is None
        if path_pid is None:
            print("Getting pids...")
            path_pid = self.dir_output / "mly_pids.csv"
            if path_pid.exists() & (update_pids == False):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            buffer = buffer, start_date = start_date, end_date = end_date, **kwargs)
        else:
            # check if the path_pid exists
            if path_pid.exists():
                print("The following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(path_pid, lat=lat, lon=lon,
                            input_csv_file=input_csv_file, input_shp_file = input_shp_file, input_place_name = input_place_name, 
                            buffer = buffer, **kwargs)
        # stop if metadata_only is True
        if metadata_only:
            print("The metadata has been downloaded")
            return
    
        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "mly_svi"
        self.panorama_output.mkdir(parents=True, exist_ok=True)
        
        # get urls
        if path_pid.exists():
            self._get_urls_mly(path_pid, resolution=resolution) 
            # download images
            self._download_images_mly(path_pid, cropped, batch_size, start_date, end_date)
        else: 
            print("There is no panorama ID to download within the given input parameters")
        
        # delete the cache directory
        if self.dir_cache.exists():
            shutil.rmtree(self.dir_cache)
            print("The cache directory has been deleted") 



import osmnx as ox

def standardize_column_names(df):
    longitude_variants = ['longitude', 'long', 'lon', 'lng', "x"]
    latitude_variants = ['latitude', 'lat', 'lt', "y"]
    # convert all column names to lowercase
    df.columns = [col.lower() for col in df.columns]      
    for col in df.columns:
        if col in longitude_variants:
            df.rename(columns={col: 'longitude'}, inplace=True)
        elif col in latitude_variants:
            df.rename(columns={col: 'latitude'}, inplace=True)
    return df

def create_buffer_gdf(gdf, buffer_distance):
    if gdf.crs == None:
        gdf = gdf.set_crs("EPSG:4326")
    
    if gdf.crs.is_projected == False:
        gdf = ox.projection.project_gdf(gdf)
    
    # Buffer the points by buffer_distance
    gdf['geometry'] = gdf.buffer(buffer_distance)
    
    # Project the GeoDataFrame back to EPSG:4326
    gdf = gdf.to_crs('EPSG:4326')
    
    return gdf

def check_and_buffer(gdf, buffer):
    # check geometry type
    geom_type = gdf.geom_type.unique()[0]
    # raise an error if the geometry is a Point/MultiPoint or a LineString/MultiLineString & buffer = 0
    if (geom_type == "Point" or geom_type == "MultiPoint" or geom_type == "LineString" or geom_type == "MultiLineString") and buffer == 0:
        raise ValueError("Buffer cannot be 0 if the geometry is either a Point/MultiPoint or a LineString/MultiLineString")
    else:
        gdf = create_buffer_gdf(gdf, buffer)
        return gdf


# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controllers.feature
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the feature extraction business logic functionalities of the Mapillary
Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Configs
from zensvi.download.mapillary.config.api.vector_tiles import VectorTiles

# Client
from zensvi.download.mapillary.models.client import Client

# Utils
from zensvi.download.mapillary.utils.verify import valid_id, points_traffic_signs_check
from zensvi.download.mapillary.utils.filter import pipeline
from zensvi.download.mapillary.utils.format import (
    merged_features_list_to_geojson,
    feature_to_geojson,
)

# Adapters
from zensvi.download.mapillary.models.api.entities import EntityAdapter

# Package imports
import mercantile
from vt2geojson.tools import vt_bytes_to_geojson


def get_feature_from_key_controller(key: int, fields: list) -> str:
    """
    A controller for getting properties of a certain image given the image key and
    the list of fields/properties to be returned

    :param key: The image key
    :type key: int

    :param fields: List of possible fields
    :type fields: list

    :return: The requested feature properties in GeoJSON format
    :rtype: str
    """

    valid_id(identity=key, image=False)

    # ? feature_to_geojson returns dict, but merged_features_list_to_geojson takes list as input
    return merged_features_list_to_geojson(
        features_list=feature_to_geojson(
            json_data=EntityAdapter().fetch_map_feature(
                map_feature_id=key, fields=fields
            )
        )
    )


def get_map_features_in_bbox_controller(
    bbox: dict,
    filter_values: list,
    filters: dict,
    layer: str = "points",
) -> str:
    """
    For extracting either map feature points or traffic signs within a bounding box

    :param bbox: Bounding box coordinates as argument
    :type bbox: dict

    :param layer: 'points' or 'traffic_signs'
    :type layer: str

    :param filter_values: a list of filter values supported by the API.
    :type filter_values: list

    :param filters: Chronological filters
    :type filters: dict

    :return: GeoJSON
    :rtype: str
    """

    # Verifying the existence of the filter kwargs
    filters = points_traffic_signs_check(filters)

    # Instantiating Client for API requests
    client = Client()

    # Getting all tiles within or intersecting the bbox
    tiles = list(
        mercantile.tiles(
            west=bbox["west"],
            south=bbox["south"],
            east=bbox["east"],
            north=bbox["north"],
            zooms=14,
        )
    )

    # Filtered features lists from different tiles will be merged into
    # filtered_features
    filtered_features = []

    for tile in tiles:
        # Decide which endpoint to send a request to based on the layer
        url = (
            VectorTiles.get_map_feature_point(x=tile.x, y=tile.y, z=tile.z)
            if layer == "points"
            else VectorTiles.get_map_feature_traffic_sign(x=tile.x, y=tile.y, z=tile.z)
        )

        res = client.get(url)

        # Decoding byte tiles
        data = vt_bytes_to_geojson(res.content, tile.x, tile.y, tile.z)

        filtered_features.extend(
            pipeline(
                data=data,
                components=[
                    # Skip filtering based on filter_values if they're not specified by the user
                    {
                        "filter": "filter_values",
                        "values": filter_values,
                        "property": "value",
                    }
                    if filter_values is not None
                    else {},
                    # Check if the features actually lie within the bbox
                    {"filter": "features_in_bounding_box", "bbox": bbox},
                    # Checks if the feature existed after a given date
                    {
                        "filter": "existed_at",
                        "existed_at": filters["existed_at"],
                    }
                    if filters["existed_at"] is not None
                    else {},
                    # Filter out all the features after a given timestamp
                    {
                        "filter": "existed_before",
                        "existed_before": filters["existed_before"],
                    }
                    if filters["existed_before"] is not None
                    else {},
                ],
            )
        )

    return merged_features_list_to_geojson(filtered_features)
